from .ExamplePage import ExamplePage


class PlugInInspector(ExamplePage):
    """Plug-in Inspector.

    Show all the public names of all registered Plug-ins.
    """

    def writeContent(self):
        wr = self.writeln
        for plugIn in self.application().plugIns():
            wr(f'<h4>{self.htmlEncode(plugIn)}</h4>')
            wr('<ul>')
            for name in dir(plugIn):
                if name.startswith('_'):
                    continue
                value = self.htmlEncode(str(getattr(plugIn, name)))
                wr(f'<li><strong>{name}</strong> = {value}</li>')
            self.writeln('</ul>')



"""Servlet factory template."""

import os
import sys
import threading

from keyword import iskeyword

from MiscUtils import AbstractError

from Servlet import Servlet

debug = False


class ServletFactory:
    """Servlet factory template.

    ServletFactory is an abstract class that defines the protocol for
    all servlet factories.

    Servlet factories are used by the Application to create servlets
    for transactions.

    A factory must inherit from this class and override uniqueness(),
    extensions() and either loadClass() or servletForTransaction().
    Do not invoke the base class methods as they all raise AbstractErrors.

    Each method is documented below.
    """

    # region Init

    def __init__(self, application):
        """Create servlet factory.

        Stores a reference to the application in self._app, because
        subclasses may or may not need to talk back to the application
        to do their work.
        """
        self._app = application
        self._imp = self._app._imp
        self._cacheClasses = self._app.setting("CacheServletClasses", True)
        self._cacheInstances = self._app.setting("CacheServletInstances", True)
        self._reloadClasses = self._app.setting("ReloadServletClasses", True)
        # All caches are keyed on the path.
        # _classCache caches the servlet classes,
        # in dictionaries with keys 'mtime' and 'class'.
        # 'mtime' is the modification time of the enclosing module.
        self._classCache = {}
        # _servletPool has lists of free reusable servlets
        self._servletPool = {}
        # _threadsafeServletCache has threadsafe servlets
        # (which are not pooled, so only one is kept at a time)
        self._threadsafeServletCache = {}
        self._importLock = threading.RLock()

    # endregion Init

    # region Info

    def name(self):
        """Return the name of the factory.

        This is a convenience for the class name.
        """
        return self.__class__.__name__

    def uniqueness(self):
        """Return uniqueness type.

        Returns a string to indicate the uniqueness of the ServletFactory's
        servlets. The Application needs to know if the servlets are unique
        per file, per extension or per application.

        Return values are 'file', 'extension' and 'application'.

        NOTE: Application so far only supports 'file' uniqueness.
        """
        raise AbstractError(self.__class__)

    def extensions(self):
        """Return a list of extensions that match this handler.

        Extensions should include the dot. An empty string indicates a file
        with no extension and is a valid value. The extension '.*' is a special
        case that is looked for a URL's extension doesn't match anything.
        """
        raise AbstractError(self.__class__)

    # endregion Info

    # region Import

    def importAsPackage(self, transaction, serverSidePathToImport):
        """Import requested module.

        Imports the module at the given path in the proper package/subpackage
        for the current request. For example, if the transaction has the URL
        http://localhost/Webware/MyContextDirectory/MySubdirectory/MyPage
        and path = 'some/random/path/MyModule.py' and the context is configured
        to have the name 'MyContext' then this function imports the module at
        that path as MyContext.MySubdirectory.MyModule . Note that the context
        name may differ from the name of the directory containing the context,
        even though they are usually the same by convention.

        Note that the module imported may have a different name from the
        servlet name specified in the URL. This is used in PSP.
        """

        # Pull out the full server side path and the context path
        request = transaction.request()
        path = request.serverSidePath()
        contextPath = request.serverSideContextPath()
        fullname = request.contextName()

        # There is no context, so import the module standalone
        # and give it a unique name:
        if not fullname or not path.startswith(contextPath):
            fullname = serverSidePathToImport
            if os.sep != '/':
                fullname = fullname.replace(os.sep, '_')
            fullname = fullname.replace('/', '_').replace('.', '_')
            name = os.path.splitext(os.path.basename(
                serverSidePathToImport))[0]
            moduleDir = os.path.dirname(serverSidePathToImport)
            module = self._importModuleFromDirectory(
                fullname, name, moduleDir, forceReload=self._reloadClasses)
            return module

        # First, we'll import the context's package.
        if os.sep != '/':
            fullname = fullname.replace(os.sep, '_')
        fullname = fullname.replace('/', '_')
        directory, contextDirName = os.path.split(contextPath)
        self._importModuleFromDirectory(
            fullname, contextDirName, directory or '.', isPackageDir=True)
        directory = contextPath

        # Now we'll break up the rest of the path into components.
        remainder = path[len(contextPath)+1:]
        if os.sep != '/':
            remainder = remainder.replace(os.sep, '/')
        remainder = remainder.split('/')

        # Import all subpackages of the context package
        for name in remainder[:-1]:
            fullname = f'{fullname}.{name}'
            self._importModuleFromDirectory(
                fullname, name, directory, isPackageDir=True)
            directory = os.path.join(directory, name)

        # Finally, import the module itself as though it was part of the
        # package or subpackage, even though it may be located somewhere else.
        moduleFileName = os.path.basename(serverSidePathToImport)
        moduleDir = os.path.dirname(serverSidePathToImport)
        name = os.path.splitext(moduleFileName)[0]
        fullname = f'{fullname}.{name}'
        module = self._importModuleFromDirectory(
            fullname, name, moduleDir, forceReload=self._reloadClasses)
        return module

    def _importModuleFromDirectory(
            self, fullModuleName, moduleName,
            directory, isPackageDir=False, forceReload=False):
        """Imports the given module from the given directory.

        fullModuleName should be the full dotted name that will be given
        to the module within Python. moduleName should be the name of the
        module in the filesystem, which may be different from the name
        given in fullModuleName. Returns the module object.
        If forceReload is True, then this reloads the module even if it
        has already been imported.

        If isPackageDir is True, then this function creates an empty
        __init__.py if that file doesn't already exist.
        """
        if debug:
            print(__file__, fullModuleName, moduleName, directory)
        module = sys.modules.get(fullModuleName)
        if module is not None and not forceReload:
            return module
        if isPackageDir:
            # check if __init__.py is in the directory
            packageDir = os.path.join(directory, moduleName)
            initPy = os.path.join(packageDir, '__init__.py')
            if (not os.path.exists(initPy) and
                    not os.path.exists(initPy + 'c')):
                # if it does not exist, make an empty one
                print(f"Creating __init__.py file at {packageDir!r}")
                try:
                    with open(initPy, 'w') as initPyFile:
                        initPyFile.write(
                            '# Auto-generated by Webware' + os.linesep)
                except Exception:
                    print("Error: __init__.py file could not be created.")
        spec = self._imp.findSpec(moduleName, directory, fullModuleName)
        module = self._imp.moduleFromSpec(spec)
        module.__donotreload__ = self._reloadClasses
        sys.modules[fullModuleName] = module
        return module

    def loadClass(self, transaction, path):
        """Load the appropriate class.

        Given a transaction and a path, load the class for creating these
        servlets. Caching, pooling, and threadsafeness are all handled by
        servletForTransaction. This method is not expected to be threadsafe.
        """
        raise AbstractError(self.__class__)

    # endregion Import

    # region Servlet Pool

    def servletForTransaction(self, transaction):
        """Return a new servlet that will handle the transaction.

        This method handles caching, and will call loadClass(trans, filepath)
        if no cache is found. Caching is generally controlled by servlets
        with the canBeReused() and canBeThreaded() methods.
        """
        request = transaction.request()
        path = request.serverSidePath()
        # Do we need to import/reimport the class
        # because the file changed on disk or isn't in cache?
        mtime = os.path.getmtime(path)
        if (path not in self._classCache
                or mtime != self._classCache[path]['mtime']):
            # Use a lock to prevent multiple simultaneous
            # imports of the same module:
            with self._importLock:
                if (path not in self._classCache
                        or mtime != self._classCache[path]['mtime']):
                    # pylint: disable = assignment-from-no-return
                    theClass = self.loadClass(transaction, path)
                    if self._cacheClasses:
                        self._classCache[path] = {
                            'mtime': mtime, 'class': theClass}
                else:
                    theClass = self._classCache[path]['class']
        else:
            theClass = self._classCache[path]['class']

        # Try to find a cached servlet of the correct class.
        # (Outdated servlets may have been returned to the pool after a new
        # class was imported, but we don't want to use an outdated servlet.)
        if path in self._threadsafeServletCache:
            servlet = self._threadsafeServletCache[path]
            if servlet.__class__ is theClass:
                return servlet
        else:
            while True:
                try:
                    servlet = self._servletPool[path].pop()
                except (KeyError, IndexError):
                    break
                else:
                    if servlet.__class__ is theClass:
                        servlet.open()
                        return servlet

        # Use a lock to prevent multiple simultaneous imports of the same
        # module. Note that (only) the import itself is already threadsafe.
        with self._importLock:
            mtime = os.path.getmtime(path)
            # pylint: disable = assignment-from-no-return
            if path not in self._classCache:
                self._classCache[path] = {
                    'mtime': mtime,
                    'class': self.loadClass(transaction, path)}
            elif mtime > self._classCache[path]['mtime']:
                self._classCache[path]['mtime'] = mtime
                self._classCache[path]['class'] = self.loadClass(
                    transaction, path)
            theClass = self._classCache[path]['class']
            if not self._cacheClasses:
                del self._classCache[path]

        # No adequate cached servlet exists, so create a new servlet instance
        servlet = theClass()
        servlet.setFactory(self)
        if servlet.canBeReused():
            if servlet.canBeThreaded():
                self._threadsafeServletCache[path] = servlet
            else:
                self._servletPool[path] = []
                servlet.open()
        return servlet

    def returnServlet(self, servlet):
        """Return servlet to the pool.

        Called by Servlet.close(), which returns the servlet
        to the servlet pool if necessary.
        """
        if (servlet.canBeReused() and not servlet.canBeThreaded()
                and self._cacheInstances):
            path = servlet.serverSidePath()
            self._servletPool[path].append(servlet)

    def flushCache(self):
        """Flush the servlet cache and start fresh.

        Servlets that are currently in the wild may find their way back
        into the cache (this may be a problem).
        """
        self._importLock.acquire()
        self._classCache = {}
        # We can't just delete all the lists, because returning
        # servlets expect it to exist.
        for key in self._servletPool:
            self._servletPool[key] = []
        self._threadsafeServletCache = {}
        self._importLock.release()

    # endregion Servlet Pool


class PythonServletFactory(ServletFactory):
    """The factory for Python servlets.

    This is the factory for ordinary Python servlets whose extensions
    are empty or .py. The servlets are unique per file since the file
    itself defines the servlet.
    """

    # region Info

    def uniqueness(self):
        return 'file'

    def extensions(self):
        # The extensions of ordinary Python servlets. Besides .py, we also
        # allow .pyc files as Python servlets, so that you can use servlets
        # in the production environment without the source code.
        # Otherwise they would be treated as ordinary files which might
        # become a security hole (though the standard configuration ignores
        # the .pyc files). If you use both extensions, make sure .py
        # comes before .pyc in the ExtensionCascadeOrder.
        return ['.py', '.pyc']

    # endregion Info

    # region Import

    def loadClass(self, transaction, path):
        # Import the module as part of the context's package
        module = self.importAsPackage(transaction, path)

        # The class name is expected to be the same as the servlet name:
        name = os.path.splitext(os.path.split(path)[1])[0]
        # Check whether such a class exists in the servlet module:
        if not hasattr(module, name):
            # If it does not exist, maybe the name has to be mangled.
            # Servlet names may have dashes or blanks in them, but classes not.
            # So we automatically translate dashes blanks to underscores:
            name = name.replace('-', '_').replace(' ', '_')
            # You may also have a servlet name that is a Python reserved word.
            # Automatically append an underscore in these cases:
            if iskeyword(name):
                name += '_'
            # If the mangled name does not exist either, report an error:
            if not hasattr(module, name):
                raise ValueError(
                    'Cannot find expected servlet class'
                    f' {name!r} in {path!r}.')
        # Pull the servlet class out of the module:
        theClass = getattr(module, name)
        if not (isinstance(theClass, object)
                and issubclass(theClass, Servlet)):
            raise TypeError(f'Expected a Servlet class: {theClass!r}')
        return theClass

    # endregion Import




"""This is the task manager Python package.

It provides a system for running any number of predefined tasks in separate
threads in an organized and controlled manner.

A task in this package is a class derived from the Task class. The task
should have a run method that, when called, performs some task.

The Scheduler class is the organizing object. It manages the addition,
execution, deletion, and well being of a number of tasks. Once you have
created your task class, you call the Scheduler to get it added to the
tasks to be run.
"""


from threading import Thread, Event
from time import localtime, time

from .TaskHandler import TaskHandler


class Scheduler(Thread):
    """The top level class of the task manager system.

    The Scheduler is a thread that handles organizing and running tasks.
    The Scheduler class should be instantiated to start a task manager session.
    Its start method should be called to start the task manager.
    Its stop method should be called to end the task manager session.
    """

    # region Init

    def __init__(self, daemon=True, exceptionHandler=None):
        Thread.__init__(self)
        self._notifyEvent = Event()
        self._nextTime = None
        self._scheduled = {}
        self._running = {}
        self._onDemand = {}
        self._isRunning = False
        self._exceptionHandler = exceptionHandler
        if daemon:
            self.setDaemon(True)

    # endregion Init

    # region Event Methods

    def wait(self, seconds=None):
        """Our own version of wait().

        When called, it waits for the specified number of seconds, or until
        it is notified that it needs to wake up, through the notify event.
        """
        try:
            self._notifyEvent.wait(seconds)
        except IOError:
            pass
        self._notifyEvent.clear()

    # endregion Event Methods

    # region Attributes

    def runningTasks(self):
        """Return all running tasks."""
        return self._running

    def running(self, name, default=None):
        """Return running task with given name.

        Returns a task with the given name from the "running" list,
        if it is present there.
        """
        return self._running.get(name, default)

    def hasRunning(self, name):
        """Check to see if a task with the given name is currently running."""
        return name in self._running

    def setRunning(self, handle):
        """Add a task to the running dictionary.

        Used internally only.
        """
        self._running[handle.name()] = handle

    def delRunning(self, name):
        """Delete a task from the running list.

        Used internally.
        """
        try:
            handle = self._running[name]
            del self._running[name]
            return handle
        except Exception:
            return None

    def scheduledTasks(self):
        """Return all scheduled tasks."""
        return self._scheduled

    def scheduled(self, name, default=None):
        """Return a task from the scheduled list."""
        return self._scheduled.get(name, default)

    def hasScheduled(self, name):
        """Checks whether task with given name is in the scheduled list."""
        return name in self._scheduled

    def setScheduled(self, handle):
        """Add the given task to the scheduled list."""
        self._scheduled[handle.name()] = handle

    def delScheduled(self, name):
        """Delete a task with the given name from the scheduled list."""
        return self._scheduled.pop(name, None)

    def onDemandTasks(self):
        """Return all on demand tasks."""
        return self._onDemand

    def onDemand(self, name, default=None):
        """Return a task from the onDemand list."""
        return self._onDemand.get(name, default)

    def hasOnDemand(self, name):
        """Checks whether task with given name is in the on demand list."""
        return name in self._onDemand

    def setOnDemand(self, handle):
        """Add the given task to the on demand list."""
        self._onDemand[handle.name()] = handle

    def delOnDemand(self, name):
        """Delete a task with the given name from the on demand list."""
        return self._onDemand.pop(name, None)

    def nextTime(self):
        """Get next execution time."""
        return self._nextTime

    def setNextTime(self, nextTime):
        """Set next execution time."""
        self._nextTime = nextTime

    def isRunning(self):
        """Check whether thread is running."""
        return self._isRunning

    # endregion Attributes

    # region Adding Tasks

    def addTimedAction(self, actionTime, task, name):
        """Add a task to be run once, at a specific time."""
        handle = self.unregisterTask(name)
        if handle:
            handle.reset(actionTime, 0, task, True)
        else:
            handle = TaskHandler(self, actionTime, 0, task, name)
        self.scheduleTask(handle)

    def addActionOnDemand(self, task, name):
        """Add a task to be run only on demand.

        Adds a task to the scheduler that will not be scheduled
        until specifically requested.
        """
        handle = self.unregisterTask(name)
        if handle:
            handle.reset(time(), 0, task, True)
        else:
            handle = TaskHandler(self, time(), 0, task, name)
        handle.setOnDemand()
        self.setOnDemand(handle)

    def addDailyAction(self, hour, minute, task, name):
        """Add an action to be run every day at a specific time.

        If a task with the given name is already registered with the
        scheduler, that task will be removed from the scheduling queue
        and registered anew as a periodic task.

        Can we make this addCalendarAction? What if we want to run
        something once a week? We probably don't need that for Webware,
        but this is a more generally useful module. This could be a
        difficult function, though. Particularly without mxDateTime.
        """
        current = localtime()
        currHour = current[3]
        currMin = current[4]

        if hour > currHour:
            hourDifference = hour - currHour
            if minute > currMin:
                minuteDifference = minute - currMin
            elif minute < currMin:
                minuteDifference = 60 - currMin + minute
                hourDifference -= 1
            else:
                minuteDifference = 0
        elif hour < currHour:
            hourDifference = 24 - currHour + hour
            if minute > currMin:
                minuteDifference = minute - currMin
            elif minute < currMin:
                minuteDifference = 60 - currMin + minute
                hourDifference -= 1
            else:
                minuteDifference = 0
        else:
            if minute > currMin:
                hourDifference = 0
                minuteDifference = minute - currMin
            elif minute < currMin:
                minuteDifference = 60 - currMin + minute
                hourDifference = 23
            else:
                hourDifference = 0
                minuteDifference = 0

        delay = (minuteDifference + (hourDifference * 60)) * 60
        self.addPeriodicAction(time() + delay, 24 * 60 * 60, task, name)

    def addPeriodicAction(self, start, period, task, name):
        """Add a task to be run periodically.

        Adds an action to be run at a specific initial time,
        and every period thereafter.

        The scheduler will not reschedule a task until the last
        scheduled instance of the task has completed.

        If a task with the given name is already registered with
        the scheduler, that task will be removed from the scheduling
        queue and registered anew as a periodic task.
        """
        handle = self.unregisterTask(name)
        if handle:
            handle.reset(start, period, task, True)
        else:
            handle = TaskHandler(self, start, period, task, name)
        self.scheduleTask(handle)

    # endregion Adding Tasks

    # region Task methods

    def unregisterTask(self, name):
        """Unregisters the named task.

        After that it can be rescheduled with different parameters,
        or simply removed.
        """

        handle = (self.delRunning(name) or
                  self.delScheduled(name) or self.delOnDemand(name))
        if handle:
            handle.unregister()
        return handle

    def runTaskNow(self, name):
        """Allow a registered task to be immediately executed.

        Returns True if the task is either currently running or was started,
        or False if the task could not be found in the list of currently
        registered tasks.
        """
        if self.hasRunning(name):
            return True
        handle = self.scheduled(name)
        if not handle:
            handle = self.onDemand(name)
        if not handle:
            return False
        self.runTask(handle)
        return True

    def demandTask(self, name):
        """Demand execution of a task.

        Allow the server to request that a task listed as being registered
        on-demand be run as soon as possible.

        If the task is currently running, it will be flagged to run again
        as soon as the current run completes.

        Returns False if the task name could not be found on the on-demand
        or currently running lists.
        """
        if self.hasRunning(name) or self.hasOnDemand(name):
            handle = self.running(name)
            if handle:
                handle.runOnCompletion()
                return True
            handle = self.onDemand(name)
            if not handle:
                return False
            self.runTask(handle)
            return True
        return False

    def stopTask(self, name):
        """Put an immediate halt to a running background task.

        Returns True if the task was either not running, or was
        running and was told to stop.
        """
        handle = self.running(name)
        if not handle:
            return False
        handle.stop()
        return True

    def stopAllTasks(self):
        """Terminate all running tasks."""
        for i in self._running:
            self.stopTask(i)

    def disableTask(self, name):
        """Specify that a task be suspended.

        Suspended tasks will not be scheduled until later enabled.
        If the task is currently running, it will not be interfered
        with, but the task will not be scheduled for execution in
        future until re-enabled.

        Returns True if the task was found and disabled.
        """
        handle = self.running(name)
        if not handle:
            handle = self.scheduled(name)
        if not handle:
            return False
        handle.disable()
        return True

    def enableTask(self, name):
        """Enable a task again.

        This method is provided to specify that a task be re-enabled
        after a suspension. A re-enabled task will be scheduled for
        execution according to its original schedule, with any runtimes
        that would have been issued during the time the task was suspended
        simply skipped.

        Returns True if the task was found and enabled.
        """
        handle = self.running(name)
        if not handle:
            handle = self.scheduled(name)
        if not handle:
            return False
        handle.enable()
        return True

    def runTask(self, handle):
        """Run a task.

        Used by the Scheduler thread's main loop to put a task in
        the scheduled hash onto the run hash.
        """
        name = handle.name()
        if self.delScheduled(name) or self.delOnDemand(name):
            self.setRunning(handle)
            handle.runTask()

    def scheduleTask(self, handle):
        """Schedule a task.

        This method takes a task that needs to be scheduled and adds it
        to the scheduler. All scheduling additions or changes are handled
        by this method. This is the only Scheduler method that can notify
        the run() method that it may need to wake up early to handle a
        newly registered task.
        """
        self.setScheduled(handle)
        if not self.nextTime() or handle.startTime() < self.nextTime():
            self.setNextTime(handle.startTime())
            self.notify()

    # endregion Task methods

    # region Misc Methods

    def notifyCompletion(self, handle):
        """Notify completion of a task.

        Used by instances of TaskHandler to let the Scheduler thread know
        when their tasks have run to completion. This method is responsible
        for rescheduling the task if it is a periodic task.
        """
        name = handle.name()
        if self.hasRunning(name):
            self.delRunning(name)
            if handle.startTime() and handle.startTime() > time():
                self.scheduleTask(handle)
            else:
                if handle.reschedule():
                    self.scheduleTask(handle)
                elif handle.isOnDemand():
                    self.setOnDemand(handle)
                    if handle.runAgain():
                        self.runTask(handle)

    def notifyFailure(self, handle):
        """Notify failure of a task.

        Used by instances of TaskHandler to let the Scheduler thread know
        if an exception has occurred within the task thread.
        """
        self.notifyCompletion(handle)
        if self._exceptionHandler is not None:
            self._exceptionHandler()

    def notify(self):
        """Wake up scheduler by sending a notify even."""
        self._notifyEvent.set()

    def start(self):
        """Start the scheduler's activity."""
        self._isRunning = True
        Thread.start(self)

    def stop(self):
        """Terminate the scheduler and its associated tasks."""
        self._isRunning = False
        self.notify()
        self.stopAllTasks()
        # Wait until the scheduler thread exits; otherwise it's possible for
        # the interpreter to exit before this thread has a chance to shut down
        # completely, which causes a traceback. Waiting 3 secs should suffice.
        self.join(3)

    # endregion Misc Methods

    # region Main Method

    def run(self):
        """The main method of the scheduler running as a background thread.

        This method is responsible for carrying out the scheduling work of
        this class on a background thread. The basic logic is to wait until
        the next action is due to run, move the task from our scheduled
        list to our running list, and run it. Other synchronized methods
        such as runTask(), scheduleTask(), and notifyCompletion(), may
        be called while this method is waiting for something to happen.
        These methods modify the data structures that run() uses to
        determine its scheduling needs.
        """
        while self._isRunning:
            if self.nextTime():
                nextTime = self.nextTime()
                currentTime = time()
                if currentTime < nextTime:
                    sleepTime = nextTime - currentTime
                    self.wait(sleepTime)
                if not self._isRunning:
                    return
                currentTime = time()
                if currentTime >= nextTime:
                    toRun = []
                    nextRun = None
                    for handle in list(self._scheduled.values()):
                        startTime = handle.startTime()
                        if startTime <= currentTime:
                            toRun.append(handle)
                        else:
                            if not nextRun:
                                nextRun = startTime
                            elif startTime < nextRun:
                                nextRun = startTime
                    self.setNextTime(nextRun)
                    for handle in toRun:
                        self.runTask(handle)
            else:
                self.wait()

    # endregion Main Method



from .ExamplePage import ExamplePage


class Introspect(ExamplePage):

    def writeContent(self):
        self.writeln('<h4>Introspection</h4>')
        self.writeln(
            "<p>The following table shows the values for various Python"
            " expressions, all of which are related to <em>introspection</em>."
            " That is to say, all the expressions examine the environment such"
            " as the object, the object's class, the module and so on.</p>")
        self.writeln('<table style="width:100%;background-color:#eef">')
        self.list('list(locals())')
        self.list('list(globals())')
        self.list('dir(self)')
        self.list('dir(self.__class__)')
        self.list('self.__class__.__bases__')
        self.list('dir(self.__class__.__bases__[0])')
        self.writeln('</table>')

    def pair(self, key, value):
        if isinstance(value, (list, tuple)):
            value = ', '.join(map(str, value))
        value = self.htmlEncode(str(value))
        self.writeln(
            '<tr style="vertical-align:top">'
            f'<td>{key}</td><td>{value}</td></tr>')

    def list(self, codeString):
        value = eval(codeString)
        if not isinstance(value, (list, tuple)):
            value = '<em>not a list or a tuple</em>'
        self.pair(codeString, value)



from SessionShelveStore import SessionShelveStore

from .TestSessionMemoryStore import TestSessionMemoryStore


class SessionShelveStoreTest(TestSessionMemoryStore):

    _storeClass = SessionShelveStore
    _storeIsOrdered = False

    def testFileShelveRestoreFiles(self):
        app = self._app
        store = self._store
        self.assertEqual(len(store), 7)
        session = store['foo-3']
        self.assertTrue('foo-0' in store and 'foo-6' in store)
        store = SessionShelveStore(
            app, restoreFiles=False, filename='Session.Store2')
        self.assertEqual(len(store), 0)
        self.assertFalse('foo-0' in store or 'foo-6' in store)
        store['foo-3'] = session
        store.storeAllSessions()
        store = SessionShelveStore(app, filename='Session.Store2')
        self.assertEqual(len(store), 1)
        self.assertTrue('foo-3' in store)
        self.assertFalse('foo-0' in store or 'foo-6' in store)



import sys
import time
from glob import glob
try:
    from cProfile import Profile
except ImportError:
    from profile import Profile

from MiscUtils.CSVParser import CSVParser


class BenchCSVParser:

    def __init__(self, profile=False, runTestSuite=True):
        self.parse = CSVParser().parse
        self._shouldProfile = profile
        self._shouldRunTestSuite = runTestSuite
        self._iterations = 1000

    def main(self):
        print('Benchmarking CSVParser ...')
        print()
        if len(sys.argv) > 1 and sys.argv[1].lower().startswith('prof'):
            self._shouldProfile = True
        if self._shouldRunTestSuite:
            print('Running test suite ...')
            from unittest import main
            main('MiscUtils.Tests.TestCSVParser', exit=False)
        start = time.time()
        if self._shouldProfile:
            prof = Profile()
            prof.runcall(self._main)
            filename = f'{self.__class__.__name__}.pstats'
            prof.dump_stats(filename)
            print('Wrote', filename)
        else:
            self._main()
        duration = time.time() - start
        print(f'{duration:.1f} secs')

    def _main(self):
        print()
        for name in glob('Sample*.csv'):
            print("Benchmark using", name, "...")
            self.benchFileNamed(name)

    def benchFileNamed(self, name):
        with open(name) as f:
            lines = f.readlines()
        for line in lines:
            for _iteration in range(self._iterations):
                # we duplicate lines to reduce the overhead of the loop
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)
                self.parse(line)


if __name__ == '__main__':
    BenchCSVParser().main()



from MiscUtils.Funcs import uniqueId
from .ExamplePage import ExamplePage


class LoginPage(ExamplePage):
    """A log-in screen for the example pages."""

    def title(self):
        return 'Log In'

    def htBodyArgs(self):
        return super().htBodyArgs() + (
            ' onload="document.loginForm.username.focus();"')

    def writeContent(self):
        self.writeln(
            '<div style="margin-left:auto;margin-right:auto;width:20em">'
            '<p>&nbsp;</p>')
        request = self.request()
        extra = request.field('extra', None)
        if (not extra and request.isSessionExpired()
                and not request.hasField('logout')):
            extra = 'You have been automatically logged out due to inactivity.'
        if extra:
            self.writeln(
                f'<p style="color:#339">{self.htmlEncode(extra)}</p>')
        if self.session().hasValue('loginId'):
            loginId = self.session().value('loginId')
        else:
            # Create a "unique" login id and put it in the form as well as
            # in the session. Login will only be allowed if they match.
            loginId = uniqueId(self)
            self.session().setValue('loginId', loginId)
        action = request.field('action', '')
        if action:
            action = f' action="{action}"'
        self.writeln(f'''<p>Please log in to view the example.
The username and password is <kbd>alice</kbd> or <kbd>bob</kbd>.</p>
<form method="post" id="loginForm"{action}>
<table style="background-color:#cce;border:1px solid #33c;width:20em">
<tr><td style="text-align:right"><label for="username">Username:</label></td>
<td><input type="text" id="username" name="username" value="admin"></td></tr>
<tr><td style="text-align:right"><label for="password">Password:</label></td>
<td><input type="password" id="password" name="password" value=""></td></tr>
<tr><td colspan="2" style="text-align:right">
<input type="submit" name="login" value="Login"></td></tr>
</table>
<input type="hidden" name="loginId" value="{loginId}">''')
        # Forward any passed in values to the user's intended page
        # after successful login, except for the special values
        # used by the login mechanism itself:
        enc = self.htmlEncode
        for name, value in request.fields().items():
            if name not in ('login', 'logout', 'loginId',
                            'username', 'password', 'extra'):
                if not isinstance(value, list):
                    value = [value]
                for v in value:
                    self.writeln('<input type="hidden"'
                                 f' name="{enc(name)}" value="{enc(v)}">')
        self.writeln('</form>\n<p>&nbsp;</p></div>')



from .SecurePage import SecurePage


class SecureCountVisits(SecurePage):
    """Secured version of counting visits example."""

    def writeContent(self):
        count = self.session().value('secure_count', 0) + 1
        self.session().setValue('secure_count', count)
        self.writeln('<h3>Counting Visits on a Secured Page</h3>')
        if self.request().isSessionExpired():
            self.writeln('<p>Your session has expired.</p>')
        self.writeln(
            "<p>You've been here"
            ' <strong style="background-color:yellow">'
            f'&nbsp;{count:d}&nbsp;</strong>'
            f" time{'' if count == 1 else 's'}.</p>")
        reload = '<a href="javascript:location.reload()">reload</a>'
        revisit = '<a href="SecureCountVisits">revisit</a>'
        self.writeln(
            '<p>This page records your visits using a session object.'
            f' Every time you {reload} or {revisit} this page,'
            ' the counter will increase. If you close your browser,'
            ' then your session will end and you will see the counter'
            ' go back to 1 on your next visit.</p>')
        self.writeln(f'<p>Try hitting {reload} now.</p>')
        user = self.loggedInUser()
        if user:
            self.writeln(
                f'<p>Authenticated user is <strong>{user}</strong>.</p>')
        revisit = '<a href="SecureCountVisits">Revisit this page</a>'
        logout = '<a href="SecureCountVisits?logout=1">Logout</a>'
        self.writeln(f'<p>{revisit} | {logout}</p>')



from Page import Page


class Servlet(Page):
    """Test of extra path info."""

    def title(self):
        return self.__doc__

    def writeBody(self):
        self.writeln('<h2>Webware Testing Servlet</h2>')
        self.writeln(f'<h3>{self.title()}</h3>')
        req = self.request()
        self.writeln(
            f"<p>serverSidePath = <code>{req.serverSidePath()}</code></p>")
        self.writeln(
            f"<p>extraURLPath = <code>{req.extraURLPath()}</code></p>")



"""This module holds the actual file writer class.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

import os

from io import StringIO


class ServletWriter:
    """This file creates the servlet source code.

    Well, it writes it out to a file at least.
    """

    _tab = '\t'
    _spaces = '    '  # 4 spaces
    _emptyString = ''

    def __init__(self, ctxt):
        self._pyFilename = ctxt.getPythonFileName()
        self._file = StringIO()
        self._tabCount = 0
        self._blockCount = 0  # a hack to handle nested blocks of python code
        self._indentSpaces = self._spaces
        self._useTabs = False
        self._useBraces = False
        self._indent = '    '
        self._userIndent = self._emptyString
        self._awakeCreated = False  # means that awake() needs to be generated

    def setIndentSpaces(self, amt):
        self._indentSpaces = ' ' * amt
        self.setIndention()

    def setIndention(self):
        self._indent = '\t' if self._useTabs else self._indentSpaces

    def setIndentType(self, indentType):
        if indentType == 'tabs':
            self._useTabs = True
            self.setIndention()
        elif indentType == 'spaces':
            self._useTabs = False
            self.setIndention()
        elif indentType == 'braces':
            self._useTabs = False
            self._useBraces = True
            self.setIndention()

    def close(self):
        pyCode = self._file.getvalue()
        self._file.close()
        if os.path.exists(self._pyFilename):
            os.remove(self._pyFilename)
        with open(self._pyFilename, 'w') as f:
            f.write(pyCode)

    def pushIndent(self):
        """this is very key, have to think more about it"""
        self._tabCount += 1

    def popIndent(self):
        if self._tabCount > 0:
            self._tabCount -= 1

    def printComment(self, start, stop, chars):
        if start and stop:
            self.println(f'## from {start}')
            self.println(f'## from {stop}')
        if chars:
            lines = chars.splitlines()
            for line in lines:
                self._file.write(self.indent(''))
                self._file.write('##')
                self._file.write(line)

    def quoteString(self, s):
        """Escape the string."""
        if s is None:
            return 'None'  # this probably won't work
        return 'r' + s

    def indent(self, s):
        """Indent the string."""
        if self._userIndent or self._tabCount > 0:
            return self._userIndent + self._indent * self._tabCount + s
        return s

    def println(self, line=None):
        """Print with indentation and a newline if none supplied."""
        if line:
            self._file.write(self.indent(line))
            if not line.endswith('\n'):
                self._file.write('\n')
        else:
            self._file.write(self.indent('\n'))

    def printChars(self, s):
        """Just prints what its given."""
        self._file.write(s)

    def printMultiLn(self, s):
        raise NotImplementedError

    def printList(self, strList):
        """Prints a list of strings with indentation and a newline."""
        for line in strList:
            self.printChars(self.indent(line))
            self.printChars('\n')

    def printIndent(self):
        """Just prints tabs."""
        self.printChars(self.indent(''))



from Page import Page

# Set this to True if you want to pass additional information.
# For security reasons, this has been disabled by default.
appInfo = False


class EditFile(Page):
    """Helper for the feature provided by the IncludeEditLink setting."""

    def writeInfo(self, key, value):
        self.writeln(f'{key}: {value}')

    def writeHTML(self):
        res = self.response()
        header = res.setHeader
        info = self.writeInfo
        req = self.request()
        env = req.environ()
        field = req.field

        header('Content-Type', 'application/x-webware-edit-file')
        header('Content-Disposition', 'inline; filename="Webware.EditFile"')

        # Basic information for editing the file:
        info('Filename', field('filename'))
        info('Line', field('line'))

        # Additional information about the hostname:
        info('Hostname', env.get('HTTP_HOST', 'localhost'))

        if appInfo:
            # Additional information about this Webware installation:
            app = self.application()
            info('ServerSidePath', app.serverSidePath())
            info('WebwarePath', app.webwarePath())



from collections import Counter

from MiscUtils.Configurable import Configurable
from .ExamplePage import ExamplePage


class DBConfig(Configurable):
    """Database configuration."""

    def defaultConfig(self):
        return {
            'dbapi': 'sqlite3',
            'database': 'demo.db',
            # 'dbapi': 'pg',
            # 'database': 'demo',
            # 'user': 'demo',
            # 'password': 'demo',
            'mincached': 5,
            'maxcached': 25
        }

    def configFilename(self):
        return 'Configs/Database.config'


# the database tables used in this example:
tables = (
    '''seminars (
        id varchar(4) primary key,
        title varchar(64) unique not null,
        cost money,
        places_left smallint)''',
    '''attendees (
        name varchar(64) not null,
        seminar varchar(4),
        paid boolean,
        primary key(name, seminar),
        foreign key (seminar) references seminars(id) on delete cascade)''')


class DBUtilsDemo(ExamplePage):
    """Example page for the DBUtils package."""

    # Initialize the buttons
    _actions = []
    _buttons = []
    for action in (
            'create tables', 'list seminars', 'list attendees',
            'add seminar', 'add attendee'):
        value = action.capitalize()
        action = action.split()
        action[1] = action[1].capitalize()
        action = ''.join(action)
        _actions.append(action)
        _buttons.append(
            f'<input name="_action_{action}" type="submit" value="{value}">')
    _buttons = tuple(_buttons)

    def initDb(self):
        """Create an instance of the DBUtils class.

        You can also do this in the contextInitialize() function
        in the __init__.py script of your application context.
        """
        app = self.application()
        try:
            dbPool = app.dbPool
        except AttributeError:
            # If a database connection pool has not yet been created,
            # create one and store it in the application as a singleton.
            config = DBConfig().config()
            if config.get('maxcached', None) is None:
                dbModName = 'persistent'
            else:
                dbModName = 'pooled'
            dbClsName = dbModName.capitalize()
            dbApiName = config.pop('dbapi', 'pg')
            if dbApiName == 'pg':  # use the PyGreSQL classic DB API
                dbModName += '_pg'
                dbClsName += 'Pg'
                if 'database' in config:
                    config['dbname'] = config['database']
                    del config['database']
                if 'password' in config:
                    config['passwd'] = config['password']
                    del config['password']
            else:  # use a DB-API 2 compliant module
                dbModName += '_db'
                dbClsName += 'DB'
            app.dbApi = app.dbPool = app.dbStatus = None
            try:
                app.dbApi = __import__(dbApiName)
                dbmod = getattr(__import__('dbutils.' + dbModName), dbModName)
                if dbApiName != 'pg':
                    config['creator'] = app.dbApi
                app.dbPool = getattr(dbmod, dbClsName)(**config)
                if hasattr(app.dbPool, 'close'):
                    app.addShutDownHandler(app.dbPool.close)
            except Exception as error:
                app.dbStatus = str(error)
            dbPool = app.dbPool
        self.dbApi = app.dbApi
        self.dbPool = dbPool
        self.dbStatus = app.dbStatus
        self.hasQuery = dbPool and dbPool.__class__.__name__.endswith('Pg')

    def title(self):
        return "DBUtils Demo"

    def actions(self):
        return super().actions() + self._actions

    def awake(self, transaction):
        super().awake(transaction)
        self.initDb()
        self._output = []

    def postAction(self, actionName):
        self.writeBody()
        del self._output
        super().postAction(actionName)

    def output(self, s):
        self._output.append(s)

    def outputMsg(self, msg, error=False):
        color = 'red' if error else 'green'
        self._output.append(f'<p style="color:{color}">{msg}</p>')

    def connection(self, shareable=True):
        if self.dbStatus:
            error = self.dbStatus
        else:
            try:
                pooledDb = self.dbPool.__class__.__name__ == 'PooledDB'
                return self.dbPool.connection(
                    shareable) if pooledDb else self.dbPool.connection()
            except self.dbApi.Error as error:
                error = str(error)
            except Exception:
                error = 'Cannot connect to the database.'
        self.outputMsg(error, True)

    def dedicatedConnection(self):
        return self.connection(False)

    def fixParamStyle(self, query):
        """Adapt parameter style to tbe used DB-API."""
        paramstyle = getattr(self.dbApi, 'paramstyle', 'format')
        if paramstyle in ('format', 'pyformat'):
            query = query.replace('?', '%s')
        return query

    def createTables(self):
        db = self.dedicatedConnection()
        if not db:
            return
        for table in tables:
            self._output.append(
                f'<p>Creating the following table:</p><pre>{table}</pre>')
            ddl = 'create table ' + table
            try:
                if self.hasQuery:
                    db.query(ddl)
                else:
                    db.cursor().execute(ddl)
                    db.commit()
            except self.dbApi.Error as error:
                if self.dbApi.__name__ != 'pg':
                    db.rollback()
                self.outputMsg(error, True)
            else:
                self.outputMsg('The table was successfully created.')
        db.close()

    def listSeminars(self):
        ids = self.request().field('id', None)
        if ids:
            if not isinstance(ids, list):
                ids = [ids]
            cmd = 'delete from seminars where id in ({})'.format(
                self.fixParamStyle(','.join('?' * len(ids))))
            db = self.dedicatedConnection()
            if not db:
                return
            try:
                if self.hasQuery:
                    db.query('begin')
                    db.query_formatted(cmd, ids)
                    db.query('end')
                else:
                    db.cursor().execute(cmd, ids)
                    db.commit()
            except self.dbApi.Error as error:
                try:
                    if self.hasQuery:
                        db.query('end')
                    else:
                        db.rollback()
                except Exception:
                    pass
                self.outputMsg(error, True)
                return
            else:
                self.outputMsg(f'Entries deleted: {len(ids)}')
        db = self.connection()
        if not db:
            return
        query = ('select id, title, cost, places_left'
                 ' from seminars order by title')
        try:
            if self.hasQuery:
                result = db.query(query).getresult()
            else:
                cursor = db.cursor()
                cursor.execute(query)
                result = cursor.fetchall()
                cursor.close()
        except self.dbApi.Error as error:
            self.outputMsg(error, True)
            return
        if not result:
            self.outputMsg('There are no seminars in the database.', True)
            return
        wr = self.output
        button = self._buttons[1].replace('List seminars', 'Delete')
        wr('<h4>List of seminars in the database:</h4>')
        wr('<form action=""><table class="NiceTable">'
           '<tr><th>ID</th><th>Seminar title</th><th>Cost</th>'
           f'<th>Places left</th><th>{button}</th></tr>')
        for seminar, title, cost, places in result:
            if places is None:
                places = 'unlimited'
            if not cost:
                cost = 'free'
            wr(f'<tr><td>{seminar}</td><td>{title}</td>'
               f'<td align="right">{cost}</td><td align="right">{places}</td>'
               '<td align="center"><input type="checkbox" name="id"'
               f' value="{seminar}"></td></tr>')
        wr('</table></form>')

    def listAttendees(self):
        ids = self.request().field('id', None)
        if ids:
            if not isinstance(ids, list):
                ids = [ids]
            fixParams = self.fixParamStyle
            places = Counter(ids)
            cmds = [('delete from attendees'
                     " where seminar||':'||name in ({})".format(
                        fixParams(','.join('?' * len(ids)))), ids),
                    *((fixParams('update seminars'
                                 ' set places_left=places_left+? where id=?'),
                       (n, i.split(':', 1)[0])) for i, n in places.items())]
            db = self.dedicatedConnection()
            if not db:
                return
            try:
                if self.hasQuery:
                    db.query('begin')
                    for cmd, values in cmds:
                        db.query_formatted(cmd, values)
                    db.query('end')
                else:
                    for cmd, values in cmds:
                        db.cursor().execute(cmd, values)
                    db.commit()
            except self.dbApi.Error as error:
                if self.hasQuery:
                    db.query('end')
                else:
                    db.rollback()
                self.outputMsg(error, True)
                return
            else:
                self.outputMsg(f'Entries deleted: {len(ids)}')
        db = self.connection()
        if not db:
            return
        query = ('select a.name, s.id, s.title,'
                 ' a.paid from attendees a,seminars s'
                 ' where s.id=a.seminar order by a.name, s.title')
        try:
            if self.hasQuery:
                result = db.query(query).getresult()
            else:
                cursor = db.cursor()
                cursor.execute(query)
                result = cursor.fetchall()
                cursor.close()
        except self.dbApi.Error as error:
            self.outputMsg(error, True)
            return
        if not result:
            self.outputMsg('There are no attendees in the database.', True)
            return
        wr = self.output
        button = self._buttons[2].replace('List attendees', 'Delete')
        wr('<h4>List of attendees in the database:</h4>')
        wr('<form action="">'
           '<table class="NiceTable">'
           '<tr><th>Name</th><th>Seminar</th><th>Paid</th>'
           f'<th>{button}</th></tr>')
        for name, seminar, title, paid in result:
            fullId = f'{seminar}:{name}'
            paid = 'Yes' if paid else 'No'
            wr(f'<tr><td>{name}</td>'
               f'<td>{title}</td>'
               f'<td align="center">{paid}</td>'
               '<td align="center">'
               f'<input type="checkbox" name="id" value="{fullId}"></td></tr>')
        wr('</table></form>')

    def addSeminar(self):
        wr = self.output
        wr('<h4>Add a seminar entry to the database:</h4>')
        wr('<form action=""><table>'
           '<tr><th>ID</th><td><input name="id" type="text"'
           ' size="4" maxlength="4"></td></tr>'
           '<tr><th>Title</th><td><input name="title" type="text"'
           ' size="40" maxlength="64"></td></tr>'
           '<tr><th>Cost</th><td><input name="cost" type="text"'
           ' size="20" maxlength="20"></td></tr>'
           '<tr><th>Places</th><td><input name="places" type="text"'
           ' size="20" maxlength="20"></td></tr>'
           f'<tr><td colspan="2" align="right">{self._buttons[3]}</td></tr>'
           '</table></form>')
        request = self.request()
        if not request.hasField('id'):
            return
        values = [request.field(name, '').strip()
                  for name in ('id', 'title', 'cost', 'places')]
        if not values[0] or not values[1]:
            self.outputMsg('You must enter a seminar ID and a title!')
            return
        if not values[2]:
            values[2] = None
        if not values[3] or not values[3].isdigit():
            values[3] = None
        db = self.dedicatedConnection()
        if not db:
            return
        cmd = self.fixParamStyle('insert into seminars values (?,?,?,?)')
        try:
            if self.hasQuery:
                db.query('begin')
                db.query_formatted(cmd, values)
                db.query('end')
            else:
                db.cursor().execute(cmd, values)
                db.commit()
        except self.dbApi.Error as error:
            if self.hasQuery:
                db.query('end')
            else:
                db.rollback()
            self.outputMsg(error, True)
        else:
            self.outputMsg(f'"{values[1]}" added to seminars.')
        db.close()

    def addAttendee(self):
        db = self.connection()
        if not db:
            return
        query = ('select id, title from seminars'
                 ' where places_left is null or places_left>0 order by title')
        try:
            if self.hasQuery:
                result = db.query(query).getresult()
            else:
                cursor = db.cursor()
                cursor.execute(query)
                result = cursor.fetchall()
                cursor.close()
        except self.dbApi.Error as error:
            self.outputMsg(error, True)
            return
        if not result:
            self.outputMsg('You have to define seminars first.')
            return
        sem = ''.join(('<select name="seminar" size="1">',
                       *(f'<option value="{id}">{title}</option>'
                         for id, title in result), '</select>'))
        wr = self.output
        wr('<h4>Add an attendee entry to the database:</h4>')
        wr('<form action=""><table>'
           '<tr><th>Name</th><td><input name="name" type="text" '
           'size="40" maxlength="64"></td></tr>'
           f'<tr><th>Seminar</th><td>{sem}</td></tr>'
           '<tr><th>Paid</th>'
           '<td><input type="radio" name="paid" value="1">Yes '
           '<input type="radio" name="paid" value="0" checked="checked">No'
           '</td></tr><tr><td colspan="2" align="right">'
           f'{self._buttons[4]}</td></tr></table></form>')
        request = self.request()
        if not request.hasField('name'):
            return
        values = []
        for name in ('name', 'seminar', 'paid'):
            values.append(request.field(name, '').strip())
        if not values[0] or not values[1]:
            self.outputMsg('You must enter a name and a seminar!')
            return
        db = self.dedicatedConnection()
        if not db:
            return
        fixParams = self.fixParamStyle
        try:
            if self.hasQuery:
                db.query('begin')
                cmd = fixParams(
                    'update seminars set places_left=places_left-1 where id=?')
                db.query_formatted(cmd, (values[1],))
                cmd = fixParams('select places_left from seminars where id=?')
                if (db.query_formatted(
                        cmd, (values[1],)).onescalar() or 0) < 0:
                    raise self.dbApi.Error("No more places left.")
                cmd = fixParams('insert into attendees values (?,?,?)')
                db.query_formatted(cmd, values)
                db.query('end')
            else:
                cursor = db.cursor()
                cmd = fixParams(
                    'update seminars set places_left=places_left-1 where id=?')
                cursor.execute(cmd, (values[1],))
                cmd = fixParams('select places_left from seminars where id=?')
                cursor.execute(cmd, (values[1],))
                if (cursor.fetchone()[0] or 0) < 0:
                    raise self.dbApi.Error("No more places left.")
                cmd = fixParams('insert into attendees values (?,?,?)')
                db.cursor().execute(cmd, values)
                cursor.close()
                db.commit()
        except self.dbApi.Error as error:
            if self.hasQuery:
                db.query('end')
            else:
                db.rollback()
            self.outputMsg(error, True)
        else:
            self.outputMsg(f'{values[0]} added to attendees.')
        db.close()

    def writeContent(self):
        wr = self.writeln
        if self._output:
            wr('\n'.join(self._output),
               '<p><a href="DBUtilsDemo">Back</a></p>')
        else:
            wr(f'<h2>Welcome to the {self.title()}!</h2>')
            wr('<p>This example shows how to use the'
               ' <a href="https://webwareforpython.github.io/DBUtils/">'
               'DBUtils</a> library with Webware for Python'
               ' as a safe and performant way to access a database.</p>')
            if self.dbPool and self.dbApi:
                wr(f'<h4>We are using {self.dbPool.__class__.__name__}'
                   f' with the {self.dbApi.__name__} database module.</h4>')
            else:
                wr('<p>No database module configured.</p>')
                wr(f'<p style="color:red">{self.dbStatus}</p>')
            wr(f'<p>Configuration: {DBConfig().config()}</p>')
            wr('<p>You can create a database configuration file to experiment'
               ' with different databases and DBUtils configurations.</p>')
            wr('<h3>Play with the demo database</h3>')
            wr('<p>This example uses a small demo database '
               'designed to track the attendees for a series of seminars '
               '(see <a href="http://www.linuxjournal.com/article/2605">"'
               'The Python DB-API"</a> by Andrew Kuchling).</p>')
            wr('<form action="">'
               '<p>{} (create the needed database tables first)</p>'
               '<p>{} {} (list all database entries)</p>'
               '<p>{} {} (add entries)</p>'
               '</form>'.format(*self._buttons))



"""Base class for tasks."""

from MiscUtils import AbstractError


class Task:
    """Abstract base class from which you have to derive your own tasks."""

    def __init__(self):
        """Subclasses should invoke super for this method."""
        # Nothing for now, but we might do something in the future.

    def run(self):
        """Override this method for you own tasks.

        Long running tasks can periodically use the proceed() method to check
        if a task should stop.
        """
        raise AbstractError(self.__class__)

    # region Utility method

    def proceed(self):
        """Check whether this task should continue running.

        Should be called periodically by long tasks to check if the system
        wants them to exit. Returns True if its OK to continue, False if
        it's time to quit.
        """
        return self._handle._isRunning

    # endregion Utility method

    # region Attributes

    def handle(self):
        """Return the task handle.

        A task is scheduled by wrapping a handler around it. It knows
        everything about the scheduling (periodicity and the like).
        Under normal circumstances you should not need the handler,
        but if you want to write period modifying run() methods,
        it is useful to have access to the handler. Use it with care.
        """
        return self._handle

    def name(self):
        """Return the unique name under which the task was scheduled."""
        return self._name

    # endregion Attributes

    # region Private method

    def _run(self, handle):
        """This is the actual run method for the Task thread.

        It is a private method which should not be overridden.
        """
        self._name = handle.name()
        self._handle = handle
        try:
            self.run()
        except Exception:
            handle.notifyFailure()
        handle.notifyCompletion()

    # endregion Private method



""""Mock Webware Transaction class."""

from .Application import Application


class Transaction:

    def __init__(self):
        self._application = Application()

    def application(self):
        return self._application



# pylint: disable=unused-import,wrong-import-order
from . import memcache  # noqa: F401 (mock memcache installation)

from SessionMemcachedStore import SessionMemcachedStore

from .TestSessionMemoryStore import TestSessionMemoryStore


class SessionMemcachedStoreTest(TestSessionMemoryStore):

    _storeClass = SessionMemcachedStore

    def setUp(self):
        TestSessionMemoryStore.setUp(self)
        self.setOnIteration()

    def tearDown(self):
        self.setOnIteration()
        TestSessionMemoryStore.tearDown(self)

    def setOnIteration(self, onIteration=None):
        self._store._onIteration = onIteration

    def testLen(self):
        self.assertEqual(len(self._store), 0)
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            len(self._store)

    def testIter(self):
        keys = list(iter(self._store))
        self.assertEqual(keys, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            list(iter(self._store))

    def testKeys(self):
        keys = self._store.keys()  # pylint: disable=assignment-from-no-return
        self.assertEqual(keys, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            self._store.keys()

    def testItems(self):
        items = self._store.items()
        self.assertEqual(items, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            self._store.items()

    def testIterItems(self):
        items = list(self._store.iteritems())
        self.assertEqual(items, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            list(self._store.iteritems())

    def testValues(self):
        values = self._store.values()
        self.assertEqual(values, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            self._store.values()

    def testIterValues(self):
        values = list(self._store.values())
        self.assertEqual(values, [])
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            self._store.values()

    def testClear(self):
        self._store.clear()
        self.setOnIteration('Error')
        with self.assertRaises(NotImplementedError):
            self._store.clear()

    def testCleanStaleSessions(self):
        self._store.cleanStaleSessions()




from time import time
from threading import Thread


class TaskHandler:
    """The task handler.

    While the Task class only knows what task to perform with the run()-method,
    the TaskHandler has all the knowledge about the periodicity of the task.
    Instances of this class are managed by the Scheduler in the scheduled,
    running and onDemand dictionaries.
    """

    # region Init

    def __init__(self, scheduler, start, period, task, name):
        self._scheduler = scheduler
        self._task = task
        self._name = name
        self._thread = None
        self._isRunning = False
        self._suspend = False
        self._lastTime = None
        self._startTime = start
        self._registerTime = time()
        self._reRegister = True
        self._reRun = False
        self._period = abs(period)
        self._isOnDemand = False

    # endregion Init

    # region Scheduling

    def reset(self, start, period, task, reRegister):
        self._startTime = start
        self._period = abs(period)
        self._task = task
        self._reRegister = reRegister

    def runTask(self):
        """Run this task in a background thread."""
        if self._suspend:
            self._scheduler.notifyCompletion(self)
            return
        self._reRun = False
        self._thread = Thread(None, self._task._run, self.name(), (self,))
        self._isRunning = True
        self._thread.start()

    def reschedule(self):
        """Determine whether this task should be rescheduled.

        Increments the startTime and returns true if this is
        a periodically executed task.
        """
        if self._period == 0 or not self._reRegister:
            return False
        if self._lastTime - self._startTime > self._period:
            # if the time taken to run the task exceeds the period
            self._startTime = self._lastTime + self._period
        else:
            self._startTime += self._period
        return True

    def notifyCompletion(self):
        self._isRunning = False
        self._lastTime = time()
        self._scheduler.notifyCompletion(self)

    def notifyFailure(self):
        self._isRunning = False
        self._lastTime = time()
        self._scheduler.notifyFailure(self)

    # endregion Scheduling

    # region Attributes

    def isRunning(self):
        return self._isRunning

    def runAgain(self):
        """Determine whether this task should be run again.

        This method lets the Scheduler check to see whether this task should be
        re-run when it terminates.
        """
        return self._reRun

    def isOnDemand(self):
        """Return True if this task is not scheduled for periodic execution."""
        return self._isOnDemand

    def setOnDemand(self, onDemand=True):
        self._isOnDemand = bool(onDemand)

    def runOnCompletion(self):
        """Request that this task be re-run after its current completion.

        Intended for on-demand tasks that are requested by the Scheduler while
        they are already running.
        """
        self._reRun = True

    def unregister(self):
        """Request that this task not be kept after its current completion.

        Used to remove a task from the scheduler.
        """
        self._reRegister = False
        self._reRun = False

    def disable(self):
        """Disable future invocations of this task."""
        self._suspend = True

    def enable(self):
        """Enable future invocations of this task."""
        self._suspend = False

    def period(self):
        """Return the period of this task."""
        return self._period

    def setPeriod(self, period):
        """Change the period for this task."""
        self._period = period

    def stop(self):
        self._isRunning = False

    def name(self):
        return self._name

    def startTime(self, newTime=None):
        if newTime:
            self._startTime = newTime
        return self._startTime

    # endregion Attributes



"""XML-RPC servlet base class

Written by Geoffrey Talvola

See Examples/XMLRPCExample.py for sample usage.
"""

import sys
import traceback
import xmlrpc.client

from RPCServlet import RPCServlet


class XMLRPCServlet(RPCServlet):
    """XMLRPCServlet is a base class for XML-RPC servlets.

    See Examples/XMLRPCExample.py for sample usage.

    For more Pythonic convenience at the cost of language independence,
    see PickleRPCServlet.
    """

    # Set to False if you do not want to allow None to be marshalled
    # as part of a response.
    allow_none = True

    def respondToPost(self, transaction):
        """Respond to a Post request.

        This is similar to the xmlrpcserver.py example from the xmlrpc
        library distribution, only it's been adapted to work within a
        Webware servlet.
        """
        try:
            # get arguments
            data = transaction.request().rawInput(rewind=1).read()
            encoding = _getXmlDeclAttr(data, "encoding")
            params, method = xmlrpc.client.loads(data)

            # generate response
            try:
                # This first test helps us to support PythonWin, which uses
                # repeated calls to __methods__.__getitem__ to determine the
                # allowed methods of an object.
                if method == '__methods__.__getitem__':
                    response = self.exposedMethods()[params[0]]
                else:
                    response = self.call(method, *params)
                if not isinstance(response, tuple):
                    response = (response,)
            except xmlrpc.client.Fault as fault:
                response = xmlrpc.client.dumps(
                    fault, encoding=encoding, allow_none=self.allow_none)
                self.sendOK('text/xml', response, transaction)
                self.handleException(transaction)
            except Exception as e:
                fault = self.resultForException(e, transaction)
                response = xmlrpc.client.dumps(
                    xmlrpc.client.Fault(1, fault),
                    encoding=encoding, allow_none=self.allow_none)
                self.sendOK('text/xml', response, transaction)
                self.handleException(transaction)
            else:
                response = xmlrpc.client.dumps(
                    response, methodresponse=1,
                    encoding=encoding, allow_none=self.allow_none)
                self.sendOK('text/xml', response, transaction)
        except Exception:
            # internal error, report as HTTP server error
            print('XMLRPCServlet internal error')
            print(''.join(traceback.format_exception(
                sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])))
            transaction.response().setStatus(500, 'Server Error')
            self.handleException(transaction)


# Helper functions:

def _getXmlDeclAttr(xml, attName):
    """Get attribute value from xml declaration (<?xml ... ?>)."""
    s = xml[6:xml.find("?>")]  # 'version = "1.0" encoding = "Cp1251"'
    p = s.find(attName)
    if p < 0:
        return None
    s = s[p + len(attName):]  # '= "Cp1251"'
    s = s[s.find('=') + 1:].strip()  # '"Cp1251"'
    return s[1:s.find(s[0], 1)]  # 'Cp1251'



"""This module tests UserManagers in different permutations."""

import os
import shutil
import unittest


class UserManagerTest(unittest.TestCase):
    """Tests for the base UserManager class."""

    def setUp(self):
        from UserKit.UserManager import UserManager
        self._mgr = UserManager()

    def testSettings(self):
        mgr = self._mgr
        value = 5.1
        mgr.setModifiedUserTimeout(value)
        self.assertEqual(mgr.modifiedUserTimeout(), value)
        mgr.setCachedUserTimeout(value)
        self.assertEqual(mgr.cachedUserTimeout(), value)
        mgr.setActiveUserTimeout(value)
        self.assertEqual(mgr.activeUserTimeout(), value)

    def testUserClass(self):
        mgr = self._mgr
        from UserKit.User import User

        class SubUser(User):
            pass

        mgr.setUserClass(SubUser)
        self.assertEqual(
            mgr.userClass(), SubUser,
            "We should be able to set a custom user class.")

        class Poser:
            pass

        # Setting a customer user class that doesn't extend UserKit.User
        # should fail.
        with self.assertRaises(TypeError):
            mgr.setUserClass(Poser)

    def tearDown(self):
        self._mgr.shutDown()
        self._mgr = None


# pylint: disable=no-member

class UserManagerToSomewhereTest:
    """Common tests for all UserManager subclasses.

    This abstract class provides some tests that all user managers should pass.
    Subclasses are responsible for overriding setUp() and tearDown() for which
    they should invoke super.
    """

    def testBasics(self):
        mgr = self._mgr
        user = self._user = mgr.createUser('foo', 'bar')
        self.assertEqual(user.manager(), mgr)
        self.assertEqual(user.name(), 'foo')
        self.assertEqual(user.password(), 'bar')
        self.assertFalse(user.isActive())
        self.assertEqual(mgr.userForSerialNum(user.serialNum()), user)
        self.assertEqual(mgr.userForExternalId(user.externalId()), user)
        self.assertEqual(mgr.userForName(user.name()), user)
        externalId = user.externalId()  # for use later in testing

        users = mgr.users()
        self.assertEqual(len(users), 1)
        self.assertEqual(
            users[0], user, 'users[0]={users[0]!r}, user={user!r}')
        self.assertEqual(len(mgr.activeUsers()), 0)
        self.assertEqual(len(mgr.inactiveUsers()), 1)

        # login
        user2 = mgr.login(user, 'bar')
        self.assertEqual(user, user2)
        self.assertTrue(user.isActive())
        self.assertEqual(len(mgr.activeUsers()), 1)
        self.assertEqual(len(mgr.inactiveUsers()), 0)

        # logout
        user.logout()
        self.assertFalse(user.isActive())
        self.assertEqual(mgr.numActiveUsers(), 0)

        # login via user
        result = user.login('bar')
        self.assertEqual(result, user)
        self.assertTrue(user.isActive())
        self.assertEqual(mgr.numActiveUsers(), 1)

        # logout via user
        user.logout()
        self.assertFalse(user.isActive())
        self.assertEqual(mgr.numActiveUsers(), 0)

        # login a 2nd time, but with bad password
        user.login('bar')
        user.login('rab')
        self.assertFalse(user.isActive())
        self.assertEqual(mgr.numActiveUsers(), 0)

        # Check that we can access the user when he is not cached
        mgr.clearCache()
        user = mgr.userForSerialNum(1)
        self.assertTrue(user)
        self.assertEqual(user.password(), 'bar')

        mgr.clearCache()
        user = self._mgr.userForExternalId(externalId)
        self.assertTrue(user)
        self.assertEqual(user.password(), 'bar')

        mgr.clearCache()
        user = self._mgr.userForName('foo')
        self.assertTrue(user)
        self.assertEqual(user.password(), 'bar')

    def testUserAccess(self):
        mgr = self._mgr
        user = mgr.createUser('foo', 'bar')

        self.assertEqual(mgr.userForSerialNum(user.serialNum()), user)
        self.assertEqual(mgr.userForExternalId(user.externalId()), user)
        self.assertEqual(mgr.userForName(user.name()), user)

        self.assertRaises(KeyError, mgr.userForSerialNum, 1000)
        self.assertRaises(KeyError, mgr.userForExternalId, 'asdf')
        self.assertRaises(KeyError, mgr.userForName, 'asdf')

        self.assertEqual(mgr.userForSerialNum(1000, 1), 1)
        self.assertEqual(mgr.userForExternalId('asdf', 1), 1)
        self.assertEqual(mgr.userForName('asdf', 1), 1)

    def testDuplicateUser(self):
        mgr = self._mgr
        user = self._user = mgr.createUser('foo', 'bar')
        self.assertEqual(user.name(), 'foo')
        self.assertRaises(KeyError, mgr.createUser, 'foo', 'bar')
        userClass = mgr.userClass()
        self.assertRaises(KeyError, userClass, mgr, 'foo', 'bar')


class UserManagerToFileTest(UserManagerToSomewhereTest, unittest.TestCase):
    """Tests for the UserManagerToFile class."""

    def setUp(self):
        from UserKit.UserManagerToFile import UserManagerToFile
        self._mgr = UserManagerToFile()
        self.setUpUserDir(self._mgr)

    def setUpUserDir(self, mgr):
        path = 'Users'
        if os.path.exists(path):
            shutil.rmtree(path, ignore_errors=True)
        os.mkdir(path)
        mgr.setUserDir(path)

    def tearDown(self):
        path = 'Users'
        if os.path.exists(path):
            shutil.rmtree(path, ignore_errors=True)


class RoleUserManagerToFileTest(UserManagerToFileTest, unittest.TestCase):
    """Tests for the RoleUserManagerToFile class."""

    def setUp(self):
        from UserKit.RoleUserManagerToFile import RoleUserManagerToFile
        self._mgr = RoleUserManagerToFile()
        self.setUpUserDir(self._mgr)



import unittest

from io import BytesIO

from WebUtils.FieldStorage import FieldStorage


class TestFieldStorage(unittest.TestCase):

    def testGetRequest(self):
        fs = FieldStorage(environ=dict(
            REQUEST_METHOD='GET', QUERY_STRING='a=1&b=2&b=3&c=3'))
        self.assertEqual(fs.getfirst('a'), '1')
        self.assertEqual(fs.getfirst('b'), '2')
        self.assertEqual(fs.getfirst('c'), '3')
        self.assertEqual(fs.getlist('a'), ['1'])
        self.assertEqual(fs.getlist('b'), ['2', '3'])
        self.assertEqual(fs.getlist('c'), ['3'])

    def testPostRequestWithQuery(self):
        fs = FieldStorage(fp=BytesIO(), environ=dict(
            REQUEST_METHOD='GET', QUERY_STRING='a=1&b=2&b=3&c=3'))
        self.assertEqual(fs.getfirst('a'), '1')
        self.assertEqual(fs.getfirst('b'), '2')
        self.assertEqual(fs.getfirst('c'), '3')
        self.assertEqual(fs.getlist('a'), ['1'])
        self.assertEqual(fs.getlist('b'), ['2', '3'])
        self.assertEqual(fs.getlist('c'), ['3'])

    def testPostRequestWithBody(self):
        fs = FieldStorage(fp=BytesIO(b'd=4&e=5&e=6&f=6'), environ=dict(
            REQUEST_METHOD='POST'))
        self.assertEqual(fs.getfirst('d'), '4')
        self.assertEqual(fs.getfirst('e'), '5')
        self.assertEqual(fs.getfirst('f'), '6')
        self.assertEqual(fs.getlist('d'), ['4'])
        self.assertEqual(fs.getlist('e'), ['5', '6'])
        self.assertEqual(fs.getlist('f'), ['6'])

    def testPostRequestOverrides(self):
        fs = FieldStorage(fp=BytesIO(b'b=8&c=9&d=4&e=5&e=6&f=6'), environ=dict(
            REQUEST_METHOD='POST', QUERY_STRING='a=1&b=2&b=3&c=3'))
        self.assertEqual(fs.getfirst('a'), '1')
        self.assertEqual(fs.getfirst('b'), '8')
        self.assertEqual(fs.getfirst('c'), '9')
        self.assertEqual(fs.getfirst('d'), '4')
        self.assertEqual(fs.getfirst('e'), '5')
        self.assertEqual(fs.getfirst('f'), '6')
        self.assertEqual(fs.getlist('a'), ['1'])
        self.assertEqual(fs.getlist('b'), ['8'])
        self.assertEqual(fs.getlist('c'), ['9'])
        self.assertEqual(fs.getlist('d'), ['4'])
        self.assertEqual(fs.getlist('e'), ['5', '6'])
        self.assertEqual(fs.getlist('f'), ['6'])

    def testPostRequestWithQueryWithSemicolon1(self):
        fs = FieldStorage(fp=BytesIO(), environ=dict(
            REQUEST_METHOD='GET', QUERY_STRING='a=1&b=2;b=3&c=3'))
        self.assertEqual(fs.getfirst('a'), '1')
        self.assertEqual(fs.getfirst('b'), '2')
        self.assertEqual(fs.getfirst('c'), '3')
        self.assertEqual(fs.getlist('a'), ['1'])
        self.assertEqual(fs.getlist('b'), ['2', '3'])
        self.assertEqual(fs.getlist('c'), ['3'])

    def testPostRequestWithQueryWithSemicolon2(self):
        fs = FieldStorage(fp=BytesIO(), environ=dict(
            REQUEST_METHOD='GET', QUERY_STRING='a=1;b=2&b=3;c=3'))
        self.assertEqual(fs.getfirst('a'), '1')
        self.assertEqual(fs.getfirst('b'), '2')
        self.assertEqual(fs.getfirst('c'), '3')
        self.assertEqual(fs.getlist('a'), ['1'])
        self.assertEqual(fs.getlist('b'), ['2', '3'])
        self.assertEqual(fs.getlist('c'), ['3'])



from .ExamplePage import ExamplePage


class OtherFileTypes(ExamplePage):

    def writeContent(self):
        self.writeln('<h4>Test for other file types:</h4>')
        self.writeLink('test.text')
        self.writeLink('test.html')

    def writeLink(self, link):
        self.writeln(f'<p><a href="Tests/{link}">{link}</a></p>')



"""HTMLTag.py

HTMLTag defines a class of the same name that represents HTML content.
An additional HTMLReader class kicks off the process of reading an HTML
file into a set of tags::

    from WebUtils.HTMLTag import HTMLReader
    reader = HTMLReader()
    tag = reader.readFileNamed('foo.html')
    tag.pprint()

Tags have attributes and children, which makes them hierarchical.
See HTMLTag class docs for more info.

Note that you imported HTMLReader instead of HTMLTag.
You only need the latter if you plan on creating tags directly.

You can discard the reader immediately if you like::

    tag = HTMLReader().readFileNamed('foo.html')

The point of reading HTML into tag objects is so that you have a concrete,
Pythonic data structure to work with. The original motivation for such a
beast was in building automated regression test suites that wanted granular,
structured access to the HTML output by the web application.

See the doc string for HTMLTag for examples of what you can do with tags.
"""

import sys

from html.parser import HTMLParser

from MiscUtils import NoDefault, AbstractError


class HTMLTagError(Exception):
    """General HTML tag error"""

    def __init__(self, msg, **values):
        Exception.__init__(self, msg)
        self.values = values.copy()


class HTMLTagAttrLookupError(HTMLTagError, LookupError):
    """HTML tag attribute lookup error"""


class HTMLTagUnbalancedError(HTMLTagError):
    """Unbalanced HTML tag error"""


class HTMLNotAllowedError(HTMLTagError):
    """HTML tag not allowed here error"""


class HTMLTagProcessingInstructionError(HTMLTagError):
    """HTML tag processing instruction error"""


class HTMLTagIncompleteError(HTMLTagError):
    """HTML tag incomplete error"""


DefaultEmptyTags = (
    'area basefont base bgsound br col colgroup frame hr'
    ' img input isindex link meta spacer wbr').split()


class HTMLTag:
    """Container class for representing HTML as tag objects.

    Tags essentially have 4 major attributes:

      * name
      * attributes
      * children
      * subtags

    Name is simple:
        print(tag.name())

    Attributes are dictionary-like in nature::

        print(tag.attr('color'))  # throws an exception if no color
        print(tag.attr('bgcolor', None))  # returns None if no bgcolor
        print(tag.attrs())

    Children are all the leaf parts of a tag, consisting of other tags
    and strings of character data::

        print(tag.numChildren())
        print(tag.childAt(0))
        print(tag.children())

    Subtags is a convenient list of only the tags in the children::

        print(tag.numSubtags())
        print(tag.subtagAt(0))
        print(tag.subtags())

    You can search a tag and all the tags it contains for a tag with
    a particular attribute matching a particular value::

        print(tag.tagWithMatchingAttr('width', '100%'))

    An HTMLTagAttrLookupError is raised if no matching tag is found.
    You can avoid this by providing a default value::

        print(tag.tagWithMatchingAttr('width', '100%', None))

    Looking for specific 'id' attributes is common in regression testing
    (it allows you to zero in on logical portions of a page),
    so a convenience method is provided::

        tag = htmlTag.tagWithId('accountTable')
    """

    # region Init and reading

    def __init__(self, name, lineNumber=None):
        if '\n' in name:
            raise ValueError('Tag name must not contain new lines')
        self._name = name
        self._attrs = {}
        self._children = []
        self._subtags = []
        self._lineNumber = lineNumber
        # Used by closedBy() and __repr__, helps with HTMLReader error messages
        self._isClosed = False

    def readAttr(self, name, value):
        """Set an attribute of the tag with the given name and value.

        A HTMLTagAttrLookupError is raised if an attribute is set twice.
        """
        attrs = self._attrs
        if name in self._attrs:
            raise HTMLTagAttrLookupError(f'name = {name!r}, attrs = {attrs!r}')
        self._attrs[name] = value

    def addChild(self, child):
        """Add a child to the receiver.

        The child will be another tag or a string (CDATA).
        """
        if not isinstance(child, (str, HTMLTag)):
            raise HTMLTagError(f'Invalid child: {child!r}')
        self._children.append(child)
        if isinstance(child, HTMLTag):
            self._subtags.append(child)

    # endregion Init and reading

    # region Access

    def name(self):
        return self._name

    def attr(self, name, default=NoDefault):
        if default is NoDefault:
            return self._attrs[name]
        return self._attrs.get(name, default)

    def hasAttr(self, name):
        return name in self._attrs

    def attrs(self):
        return self._attrs

    def numAttrs(self):
        return len(self._attrs)

    def childAt(self, index):
        return self._children[index]

    def numChildren(self):
        return len(self._children)

    def children(self):
        return self._children

    def subtagAt(self, index):
        return self._subtags[index]

    def numSubtags(self):
        return len(self._subtags)

    def subtags(self):
        return self._subtags

    # endregion Access

    # region Printing

    def pprint(self, out=None, indent=0):
        if out is None:
            out = sys.stdout
        wr = out.write
        spacer = ' '*4*indent
        wr(f'{spacer}<{self._name}>\n')
        for key, value in self._attrs.items():
            wr(f'{spacer}  {key.ljust(12)} = {value}\n')
        indent += 1
        for child in self._children:
            if isinstance(child, HTMLTag):
                child.pprint(out, indent)
            else:
                wr(f'{spacer}    {child}\n')
        wr(f'{spacer}</{self._name}>\n')

    def __repr__(self):
        r = ['<', self._name]
        attrs = self._attrs
        if attrs:
            for key in sorted(attrs):
                r.extend([' ', key, '="', attrs[key], '"'])
        r.append('>')
        if self._lineNumber or self._isClosed:
            r.append(' (')
            if self._lineNumber:
                r.append(str(self._lineNumber))
            if self._isClosed:
                if self._lineNumber:
                    r.append('; ')
                r.append(f'closed by {self._closedBy} at {self._closedAt}')
            r.append(')')
        r = ''.join(r)
        return r

    # endregion Printing

    # region Searching

    def tagWithMatchingAttr(self, name, value, default=NoDefault):
        """Search for tag with matching attributes.

        Performs a depth-first search for a tag with an attribute that matches
        the given value. If the tag cannot be found, a KeyError will be raised
        *unless* a default value was specified, which is then returned.

        Example::

            tag = tag.tagWithMatchingAttr('bgcolor', '#FFFFFF', None)
        """
        tag = self._tagWithMatchingAttr(name, value)
        if tag is None:
            if default is NoDefault:
                raise HTMLTagAttrLookupError(
                    'name = {name!r}, value = {value!r}')
            return default
        return tag

    def tagWithId(self, id_, default=NoDefault):
        """Search for tag with a given id.

        Finds and returns the tag with the given id. As in::

            <td id=foo> bar </td>

        This is just a cover for::

            tagWithMatchingAttr('id', id_, default)

        But searching for id's is so popular (at least in regression testing
        web sites) that this convenience method is provided.
        Why is it so popular? Because by attaching ids to logical portions
        of your HTML, your regression test suite can quickly zero in on them
        for examination.
        """
        return self.tagWithMatchingAttr('id', id_, default)

    # endregion Printing

    # region Parsing (HTMLReader)

    def closedBy(self, name, lineNumber):

        self._isClosed = True
        self._closedBy = name
        self._closedAt = lineNumber

    # endregion Parsing (HTMLReader)

    # region Self utility

    def _tagWithMatchingAttr(self, name, value):
        """Search for tag with matching attributes.

        Performs a depth-first search for a tag with an attribute that matches
        the given value. Returns None if the tag cannot be found. The method
        `tagWithMatchingAttr()` (e.g., sans underscore) is more commonly used.
        """
        if self._attrs.get(name) == value:
            return self
        for tag in self._subtags:
            matchingTag = tag._tagWithMatchingAttr(name, value)
            if matchingTag:
                return matchingTag
        return None

    # endregion Self utility


class HTMLReader(HTMLParser):
    """Reader class for representing HTML as tag objects.

    NOTES

    * Special attention is required regarding tags like ``<p>`` and ``<li>``
      which sometimes are closed and sometimes not. HTMLReader can deal with
      both situations (closed and not) provided that:

      * the file doesn't change conventions for a given tag
      * the reader knows ahead of time what to expect

    Be default, `HTMLReader` assumes that ``<p>`` and ``<li>`` will be closed
    with ``</p>`` and ``</li>`` as the official HTML spec encourages.

    But if your files don't close certain tags that are supposed to be
    required, you can do this::

        HTMLReader(extraEmptyTags=['p', 'li'])

    or::

        reader.extendEmptyTags(['p', 'li'])

    or just set them entirely::

        HTMLReader(emptyTags=['br', 'hr', 'p'])
        reader.setEmptyTags(['br', 'hr', 'p'])

    Although there are quite a few. Consider the DefaultEmptyTags global
    list (which is used to initialize the reader's tags) which contains
    about 16 tag names.

    If an HTML file doesn't conform to the reader's expectation, you will get
    an exception (see more below for details).

    If your HTML file doesn't contain root ``<html> ... </html>`` tags wrapping
    everything, a fake root tag will be constructed for you, unless you pass
    in fakeRootTagIfNeeded=False.

    Besides fixing your reader manually, you could conceivably loop through
    the permutations of the various empty tags to see if one of them resulted
    in a correct read.

    Or you could fix the HTML.

      * The reader ignores extra preceding and trailing whitespace by stripping
        it from strings. I suppose this is a little harsher than reducing spans
        of preceding and trailing whitespace down to one space, which is what
        really happens in an HTML browser.

      * The reader will not read past the closing ``</html>`` tag.

      * The reader is picky about the correctness of the HTML you feed it.
        If tags are not closed, overlap (instead of nest) or left unfinished,
        an exception is thrown. These include `HTMLTagUnbalancedError`,
        `HTMLTagIncompleteError` and `HTMLNotAllowedError` which all inherit
        `HTMLTagError`.

        This pickiness can be quite useful for the validation of the HTML of
        your own applications.
    """

    # region Init

    def __init__(
            self, emptyTags=None, extraEmptyTags=None,
            fakeRootTagIfNeeded=True):
        super().__init__()
        self._filename = None
        self._rootTag = None
        self._fakeRootTagIfNeeded = fakeRootTagIfNeeded
        self._usedFakeRootTag = False
        self._tagStack = []
        self._finished = False

        # Options
        self._printsStack = False
        self._ignoreWS = True
        self._endingTag = 'html'

        # Handle optional args
        if emptyTags is not None:
            self.setEmptyTags(emptyTags)
        else:
            self.setEmptyTags(DefaultEmptyTags)
        if extraEmptyTags is not None:
            self.extendEmptyTags(extraEmptyTags)

    # endregion Init

    # region Reading

    def readFileNamed(self, filename, retainRootTag=True):
        """Read the given file.

        Relies on readString(). See that method for more information.
        """
        self._filename = filename
        with open(filename) as f:
            contents = f.read()
        return self.readString(contents, retainRootTag)

    def readString(self, string, retainRootTag=True):
        """Read the given string, store the results and return the root tag.

        You could continue to use HTMLReader object or disregard it and simply
        use the root tag.
        """
        self._rootTag = None
        self._tagStack = []
        self._finished = False
        self.reset()
        self._lineNumber = 1
        self.computeTagContainmentConfig()
        try:
            for line in string.splitlines():
                self.feed(line + '\n')
                self._lineNumber += 1
            self.close()
        finally:
            self.reset()
        if retainRootTag:
            return self._rootTag
        tag = self._rootTag
        self._rootTag = None
        return tag

    # endregion Reading

    # region Printing

    def pprint(self, out=None):
        """Pretty prints the tag, its attributes and all its children.

        Indentation is used for subtags.
        Print 'Empty.' if there is no root tag.
        """
        if self._rootTag:
            self._rootTag.pprint(out)
        else:
            if out is None:
                out = sys.stdout
            out.write('Empty.')

    # endregion Printing

    # region Access

    def rootTag(self):
        """Return the root tag.

        May return None if no HTML has been read yet, or if the last
        invocation of one of the read methods was passed retainRootTag=False.
        """
        return self._rootTag

    def filename(self):
        """Return the name of the file if one has been read, otherwise None."""
        return self._filename

    def emptyTags(self):
        """Return a list of empty tags.

        See also: class docs and setEmptyTags().
        """
        return self._emptyTagList

    def setEmptyTags(self, tagList):
        """Set the HTML tags that are considered empty such as <br> and <hr>.

        The default is found in the global, DefaultEmptyTags, and is fairly
        thorough, but does not include <p>, <li> and some other tags that
        HTML authors often use as empty tags.
        """
        self._emptyTagList = list(tagList)
        self._updateEmptyTagSet()

    def extendEmptyTags(self, tagList):
        """Extend the current list of empty tags with the given list."""
        self._emptyTagList.extend(tagList)
        self._updateEmptyTagSet()

    # endregion Access

    # region Debugging

    def error(self, message):
        raise RuntimeError(f'HTMLReader error: {message}')

    def printsStack(self):
        return self._printsStack

    def setPrintsStack(self, flag):
        """Set the boolean value of the "prints stack" option.

        This is a debugging option which will print the internal tag stack
        during HTML processing. The default value is False.
        """
        self._printsStack = flag

    # endregion Debugging

    # region Command line

    def main(self, args=None):
        """The command line equivalent of readFileNamed().

        Invoked when HTMLTag is run as a program.
        """
        if args is None:
            args = sys.argv
        if len(args) < 2:
            self.usage()
        return self.readFileNamed(args[1])

    def usage(self):
        print('HTMLTag: usage:  HTMLTag <html file>')
        sys.exit(1)

    # endregion Command line

    # region HTMLParser handlers

    def handle_data(self, data):
        if self._finished:
            return
        if not isinstance(data, str):
            raise HTMLTagError(f'Encountered non-string data: {data!r}')
        if self._ignoreWS:
            data = data.strip()
            if not data:
                return
        if self._tagStack:
            self._tagStack[-1].addChild(data)
        else:
            print(f'<> data={data!r}')

    def handle_pi(self, data):
        raise HTMLTagProcessingInstructionError(
            f'Was not expecting a processing instruction: {data!r}')

    def handle_starttag(self, name, attrs):  # pylint: disable=arguments-differ
        if self._finished:
            return
        tag = HTMLTag(name, lineNumber=self._lineNumber)
        for attrName, value in attrs:
            tag.readAttr(attrName, value)
        if name in self._emptyTagSet:
            # We'll never have any children. Boo hoo.
            if not self._rootTag:
                raise HTMLTagError(
                    f'Cannot start HTML with an empty tag: {tag!r}')
            self._tagStack[-1].addChild(tag)
            empty = True
        else:
            # We could have children, so we go on the stack
            # Also, if this is the first tag, then make it the root.
            # If it's the first tag and it isn't an <html> tag,
            # create a fake "container" html tag.
            if self._tagStack:
                lastTag = self._tagStack[-1]
                # is this legal?
                tagConfig = self._tagContainmentConfig.get(lastTag.name())
                if tagConfig:
                    tagConfig.encounteredTag(name, self._lineNumber)
                # tell last tag about his new child
                lastTag.addChild(tag)
            elif name != 'html' and self._fakeRootTagIfNeeded:
                self._rootTag = HTMLTag('html')
                self._tagStack.append(self._rootTag)
                self._tagStack[-1].addChild(tag)
                self._usedFakeRootTag = True
            else:
                self._rootTag = tag
            self._tagStack.append(tag)
            empty = False
        if self._printsStack:
            prefix = ('START', '-----')[empty]
            print(f'{prefix} {name.ljust(6)}: {self._tagStack!r}')

    def handle_endtag(self, name):  # pylint: disable=arguments-differ
        if self._finished:
            return
        if name == self._endingTag:
            self._finished = True
        openingTag = self._tagStack.pop()
        if self._printsStack:
            print(f'END   {name.ljust(6)}: {self._tagStack}')
        if openingTag.name() != name:
            raise HTMLTagUnbalancedError(
                f'line {self._lineNumber}:'
                f' opening is {openingTag}, but closing is <{name}>.',
                line=self._lineNumber, opening=openingTag.name(),
                closing=name, tagStack=self._tagStack)
        openingTag.closedBy(name, self._lineNumber)

    def close(self):
        stackSize = len(self._tagStack)
        if stackSize > 0 and not (stackSize == 1 and self._usedFakeRootTag):
            raise HTMLTagIncompleteError(
                f'line {self._lineNumber}: tagStack = {self._tagStack!r}',
                line=self._lineNumber, tagStack=repr(self._tagStack))
        super().close()

    # endregion HTMLParser handlers

    # region Self utility

    def _updateEmptyTagSet(self):
        """Create a set out of the empty tag list for quick look up."""
        self._emptyTagSet = set(self._emptyTagList)

    # The following dict defines for various tags either:
    #    + the complete set of tags that can be contained within
    #    - a set of tags that cannot be contained within
    # This information helps HTMLReader detect some types of errors
    # earlier and other types of errors, it would never detect.
    tagContainmentConfig = {
        'html':   'canOnlyHave head body',
        'head':   'cannotHave  html head body',
        'body':   'cannotHave  html head body',
        'table':  'canOnlyHave tr thead tbody tfoot a',
        'tr':     'canOnlyHave th td',
        'td':     'cannotHave  td tr',
        'select': 'canOnlyHave option',
    }

    def computeTagContainmentConfig(self):
        config = {}
        for key, value in self.tagContainmentConfig.items():
            if isinstance(value, str):
                value = value.split()
                configClass = configClassForName.get(value.pop(0))
                if configClass is None:
                    raise KeyError(
                        f'Unknown config name {key!r} for value {value!r}'
                        f' in {self.__class__.__name__}.tagContainmentConfig')
                config[key] = configClass(key, value)
            else:
                if not isinstance(value, TagConfig):
                    raise TypeError(f'key={key!r}, value={value!r}')
                config[key] = value
        self._tagContainmentConfig = config

    # endregion Self utility


class TagConfig:

    def __init__(self, name, tags):
        self.name = name
        # turn tag list into a set for fast lookup (avoid linear searches)
        self.tags = set(tags)

    def encounteredTag(self, tag, lineNum):
        raise AbstractError(self.__class__)


class TagCanOnlyHaveConfig(TagConfig):

    def encounteredTag(self, tag, lineNum):
        if tag.lower() not in self.tags:
            raise HTMLNotAllowedError(
                f'line {lineNum}: the tag {tag!r} is not allowed in'
                f' {self.name!r} which can only have {sorted(self.tags)!r}.',
                line=lineNum, encounteredTag=tag, containingTag=self.name,
                canOnlyHave=self.tags)


class TagCannotHaveConfig(TagConfig):

    def encounteredTag(self, tag, lineNum):
        if tag.lower() in self.tags:
            raise HTMLNotAllowedError(
                f'line {lineNum}: The tag {tag!r} is not allowed in'
                ' {self.name!r} which cannot have {sorted(self.tags!r}.',
                line=lineNum, enounteredTag=tag, containingTag=self.name,
                cannotHave=self.tags)


configClassForName = dict(
    canOnlyHave=TagCanOnlyHaveConfig, cannotHave=TagCannotHaveConfig)


if __name__ == '__main__':
    html = HTMLReader().main()
    html.pprint()



"""Webware for Python"""


def addToSearchPath():
    """Add the Webware package to the search path for Python modules."""
    import sys
    webwarePath = __path__[0]
    if webwarePath not in sys.path:
        sys.path.insert(0, webwarePath)


def mockAppWithPlugins(path=None, settings=None, development=None):
    """Return a mock application with all plugins loaded."""
    addToSearchPath()
    from MockApplication import MockApplication
    from PlugInLoader import PlugInLoader
    app = MockApplication(path, settings, development)
    PlugInLoader(app).loadPlugIns()
    return app



#!/usr/bin/env python3

"""WSGI script to be used with Webware for Python."""

import os
import sys

import webware

libDirs = []
workDir = None
development = None
settings = {}

webware.addToSearchPath()

for libDir in reversed(libDirs):
    if libDir != '.' and libDir not in sys.path:
        sys.path.insert(0, libDir)

if workDir is None:
    workDir = os.path.dirname(os.path.dirname(__file__))

if workDir:
    os.chdir(workDir)

if '.' in libDirs:
    sys.path.insert(0, workDir)

from Application import Application

application = Application(
    workDir, settings=settings, development=development)



"""Test Webware Mock Environment"""

import io
import sys
import unittest

import webware


class TestMocking(unittest.TestCase):

    def setUp(self):
        self.sysPath = sys.path
        self.webwarePath = webware.__path__[0]
        sys.path = [path for path in sys.path if path != self.webwarePath]
        self.sysStdout = sys.stdout
        sys.stdout = io.StringIO()
        self.getOutput = sys.stdout.getvalue

    def tearDown(self):
        sys.stdout = self.sysStdout
        sys.path = self.sysPath

    def testAddToSearchPath(self):
        assert self.webwarePath not in sys.path
        webware.addToSearchPath()
        assert self.webwarePath in sys.path

    def testLoadPlugins(self):
        webware.addToSearchPath()
        app = webware.mockAppWithPlugins(
            self.webwarePath, settings={'PrintPlugIns': True})
        self.assertEqual(app.__class__.__name__, 'MockApplication')
        output = self.getOutput().splitlines()
        output = [line.split(' at ', 1)[0] for line in output]
        self.assertEqual(output, [
            'Plug-ins list: MiscUtils, WebUtils, TaskKit, UserKit, PSP',
            'Loading plug-in: MiscUtils', 'Loading plug-in: WebUtils',
            'Loading plug-in: TaskKit', 'Loading plug-in: UserKit',
            'Loading plug-in: PSP', ''])



"""DBPool.py

Implements a pool of cached connections to a database for any DB-API 2
compliant database module. This should result in a speedup for persistent
applications like Webware. The pool of connections is threadsafe regardless
of whether the used DB-API 2 general has a threadsafety of 1 or 2.

For more information on the DB-API 2 specification, see `PEP 249`_.

.. _PEP 249: https://www.python.org/dev/peps/pep-0249/

The idea behind DBPool is that it's completely seamless, so once you have
established your connection, use it just as you would any other DB-API 2
compliant module. For example::

    import pgdb  # import used DB-API 2 module
    from MiscUtils.DBPool import DBPool
    dbpool = DBPool(pgdb, 5, host=..., database=..., user=..., ...)
    db = dbpool.connection()

Now use "db" exactly as if it were a pgdb connection. It's really
just a proxy class.

db.close() will return the connection to the pool, not actually
close it. This is so your existing code works nicely.

DBPool is actually intended to be a demonstration of concept not to be
used in a productive environment. It is really a very simple solution with
several drawbacks. For instance, pooled database connections which have
become invalid are not automatically recovered. For a more sophisticated
solution, please have a look at the DBUtils_ package.

.. _DBUtils: https://webwareforpython.github.io/dbutils/


CREDIT

  * Contributed by Dan Green.
  * Thread safety bug found by Tom Schwaller.
  * Fixes by Geoff Talvola (thread safety in `_threadsafe_get_connection()`).
  * Clean up by Chuck Esterbrook.
  * Fix unthreadsafe functions which were leaking, Jay Love.
  * Eli Green's webware-discuss comments were lifted for additional docs.
  * Coding and comment clean-up by Christoph Zwerschke.
"""


class DBPoolError(Exception):
    """General database pooling error."""


class NotSupportedError(DBPoolError):
    """Missing support from database module error."""


class PooledConnection:
    """A wrapper for database connections to help with DBPool.

    You don't normally deal with this class directly,
    but use DBPool to get new connections.
    """

    def __init__(self, pool, con):
        self._con = con
        self._pool = pool

    def close(self):
        """Close the pooled connection."""
        # Instead of actually closing the connection,
        # return it to the pool so it can be reused.
        if self._con is not None:
            self._pool.returnConnection(self._con)
            self._con = None

    def __getattr__(self, name):
        # All other members are the same.
        return getattr(self._con, name)

    def __del__(self):
        self.close()


class DBPool:

    def __init__(self, dbapi, maxconnections, *args, **kwargs):
        """Set up the database connection pool.

        `dbapi`:
          the DB-API 2 compliant module you want to use
        `maxconnections`:
          the number of connections cached in the pool
        `args`, `kwargs`:
          the parameters that shall be used to establish the database
          connections using ``dbapi.connect()``
        """
        try:
            threadsafety = dbapi.threadsafety
        except Exception:
            threadsafety = None
        if threadsafety == 0:
            raise NotSupportedError(
                "Database module does not support any level of threading.")
        if dbapi.threadsafety == 1:
            # If there is no connection level safety, build
            # the pool using the synchronized queue class
            # that implements all the required locking semantics.
            from queue import Queue
            self._queue = Queue(maxconnections)  # create the queue
            self.connection = self._unthreadsafeGetConnection
            self.addConnection = self._unthreadsafeAddConnection
            self.returnConnection = self._unthreadsafeReturnConnection
        elif dbapi.threadsafety in (2, 3):
            # If there is connection level safety, implement the
            # pool with an ordinary list used as a circular buffer.
            # We only need a minimum of locking in this case.
            from threading import Lock
            self._lock = Lock()  # create a lock object to be used later
            self._nextCon = 0  # the index of the next connection to be used
            self._connections = []  # the list of connections
            self.connection = self._threadsafeGetConnection
            self.addConnection = self._threadsafeAddConnection
            self.returnConnection = self._threadsafeReturnConnection
        else:
            raise NotSupportedError(
                'Database module threading support cannot be determined.')
        # Establish all database connections (it would be better to
        # only establish a part of them now, and the rest on demand).
        for _count in range(maxconnections):
            self.addConnection(dbapi.connect(*args, **kwargs))

    # The following functions are used with DB-API 2 modules
    # that do not have connection level threadsafety, like PyGreSQL.
    # However, the module must be threadsafe at the module level.
    # Note: threadsafe/unthreadsafe refers to the DB-API 2 module,
    # not to this class which should be threadsafe in any case.

    def _unthreadsafeGetConnection(self):
        """Get a connection from the pool."""
        return PooledConnection(self, self._queue.get())

    def _unthreadsafeAddConnection(self, con):
        """Add a connection to the pool."""
        self._queue.put(con)

    def _unthreadsafeReturnConnection(self, con):
        """Return a connection to the pool.

        In this case, the connections need to be put
        back into the queue after they have been used.
        This is done automatically when the connection is closed
        and should never be called explicitly outside of this module.
        """
        self._unthreadsafeAddConnection(con)

    # The following functions are used with DB-API 2 modules
    # that are threadsafe at the connection level, like psycopg.
    # Note: In this case, connections are shared between threads.
    # This may lead to problems if you use transactions.

    def _threadsafeGetConnection(self):
        """Get a connection from the pool."""
        with self._lock:
            nextCon = self._nextCon
            con = PooledConnection(self, self._connections[nextCon])
            nextCon += 1
            if nextCon >= len(self._connections):
                nextCon = 0
            self._nextCon = nextCon
            return con

    def _threadsafeAddConnection(self, con):
        """"Add a connection to the pool."""
        self._connections.append(con)

    def _threadsafeReturnConnection(self, con):
        """Return a connection to the pool.

        In this case, the connections always stay in the pool,
        so there is no need to do anything here.
        """
        # does nothing



"""Webware page template class for pages with a sidebar."""

from Page import Page


class SidebarPage(Page):
    """Webware page template class for pages with a sidebar.

    SidebarPage is an abstract superclass for pages that have a sidebar
    (as well as a header and "content well"). Sidebars are normally used
    for navigation (e.g., a menu or list of links), showing small bits
    of info and occasionally a simple form (such as login or search).

    Subclasses should override cornerTitle(), writeSidebar() and
    writeContent() (and title() if necessary; see Page).

    The utility methods menuHeading() and menuItem() can be used by
    subclasses, typically in their implementation of writeSidebar().

    Webware itself uses this class: Examples/ExamplePage and Admin/AdminPage
    both inherit from it.
    """

    # region StyleSheet

    _styleSheet = '''
html {
  height: 100%;
}
body {
  background-color: white;
  color: #080810;
  font-size: 11pt;
  font-family: Tahoma,Verdana,Arial,Helvetica,sans-serif;
  margin: 0;
  padding: 0;
  height: 100%;
}
h1 { font-size: 18pt; }
h2 { font-size: 16pt; }
h3 { font-size: 14pt; }
h4 { font-size: 12pt; }
h5 { font-size: 11pt; }
#Page {
  display: table;
  width: 100%;
  height: 100%;
}
#Page > div {
  display: table-row;
}
#Page > div > div {
  display: table-cell;
}
#CornerTitle {
  background-color: #101040;
  color: white;
  padding: 6pt 6pt;
  font-size: 14pt;
  text-align: center;
  vertical-align: middle;
}
#Banner {
  background-color: #202080;
  color: white;
  padding: 8pt 6pt;
  font-size: 16pt;
  font-weight: bold;
  text-align: center;
  vertical-align: middle;
}
#Sidebar {
    background-color: #e8e8f0;
    padding: 4pt;
    font-size: 10pt;
    line-height: 13pt;
    vertical-align: top;
    white-space: nowrap;
    height: 100%;
}
#Sidebar div {
    margin-bottom: 1pt;
}
#Sidebar div.MenuHeading {
    font-weight: bold;
    margin-top: 6pt;
    margin-bottom: 3pt;
    width: 12em;
}
#Content {
    padding: 8pt;
    vertical-align: top;
    width: 100%;
}
table.NiceTable {
    margin-bottom: 4pt;
}
table.NiceTable td, .Data {
    background-color: #eee;
    color: #111;
}
table.NiceTable th {
    text-align: left;
}
table.NiceTable tr.TopHeading th {
    text-align: center;
}
table tr th, .SubHeading {
    background-color: #ccc;
    color: black;
}
table tr.TopHeading th, .TopHeading {
background-color: #555;
color: white;
}
#Content table tr.NoTable td {
    background-color: white;
    color: #080810;
}
table.NiceTable th a:link, table.NiceTable th a:visited {
    color: #101040;
}
'''

    def writeStyleSheet(self):
        """We're using a simple internal style sheet.

        This way we avoid having to care about where an external style
        sheet should be located when this class is used in another context.
        """
        self.writeln(f'<style>{self._styleSheet}</style>')

    # endregion StyleSheet

    # region Content methods

    def writeBodyParts(self):
        wr = self.writeln
        wr('<div id="Page"><div>')
        self.writeBanner()
        wr('</div><div><div id="Sidebar">')
        self.writeSidebar()
        wr('</div><div id="Content">')
        self.writeContent()
        wr('</div></div></div>')

    def writeBanner(self):
        self.writeln(
            f'<div id="CornerTitle">{self.cornerTitle()}</div>',
            f'<div id="Banner">{self.title()}</div>')

    def writeSidebar(self):
        self.writeWebwareSidebarSections()

    @staticmethod
    def cornerTitle():
        return ''

    # endregion Content methods

    # region Menu

    def menuHeading(self, title):
        self.writeln(f'<div class="MenuHeading">{title}</div>')

    def menuItem(self, title, url=None, suffix=None, indentLevel=1):
        if url is not None:
            title = f'<a href="{url}">{title}</a>'
        if suffix:
            title += ' ' + suffix
        margin = 4 * indentLevel
        self.writeln(f'<div style="margin-left:{margin}pt">{title}</div>')

    # endregion Menu

    # region Sidebar

    def writeWebwareSidebarSections(self):
        """Write sidebar sections.

        This method (and consequently the methods it invokes)
        are provided for Webware's example and admin pages.
        It writes sections such as contexts, e-mails, exits and versions.
        """
        self.writeContextsMenu()
        self.writeWebwareEmailMenu()
        self.writeWebwareExitsMenu()
        self.writeVersions()

    def writeContextsMenu(self):
        self.menuHeading('Contexts')
        servletPath = self.request().servletPath()
        contexts = sorted(
            ctx for ctx in self.application().contexts()
            if ctx != 'default' and '/' not in ctx)
        for context in contexts:
            self.menuItem(context, f'{servletPath}/{context}/')

    def writeWebwareEmailMenu(self):
        self.menuHeading('E-mail')
        self.menuItem(
            'webware-discuss', 'mailto:webware-discuss@lists.sourceforge.net')

    def writeWebwareExitsMenu(self):
        self.menuHeading('Exits')
        self.menuItem('Webware', 'https://webwareforpython.github.io/w4py3/')
        self.menuItem('Python', 'https://www.python.org/')

    def writeVersions(self):
        app = self.application()
        self.menuHeading('Versions')
        self.menuItem('Webware ' + app.webwareVersionString())
        from sys import version
        self.menuItem('Python ' + version.split(None, 1)[0])

    def writeContent(self):
        self.writeln('Woops, someone forgot to override writeContent().')

    # endregion Sidebar



"""Session store using files."""

import os
import threading

from MiscUtils import NoDefault

from SessionStore import SessionStore


class SessionFileStore(SessionStore):
    """A session file store.

    Stores the sessions on disk in the Sessions/ directory,
    one file per session.
    """

    _extension = '.ses'

    # region Init

    def __init__(self, app, restoreFiles=None):
        """Initialize the session file store.

        If restoreFiles is true, and sessions have been saved to file,
        the store will be initialized from these files.
        """
        SessionStore.__init__(self, app)
        if restoreFiles is None:
            restoreFiles = self._retain
        self._sessionDir = app._sessionDir
        self._lock = threading.RLock()
        if not restoreFiles:
            self.clear()

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions in the store."""
        return len(self.keys())

    def __getitem__(self, key):
        """Get a session item, loading it from the session file."""
        filename = self.filenameForKey(key)
        with self._lock:
            try:
                sessionFile = open(filename, 'rb')
            except IOError as e:
                raise KeyError(key) from e  # session file not found
            try:
                try:
                    value = self.decoder()(sessionFile)
                finally:
                    sessionFile.close()
            except Exception as e:
                print("Error loading session from disk:", key)
                self.application().handleException()
                try:  # remove the session file because it is corrupt
                    os.remove(filename)
                except Exception:
                    pass
                raise KeyError(key) from e
        return value

    def __setitem__(self, key, value):
        """Set a session item, saving it to a session file."""
        dirty = value.isDirty()
        if self._alwaysSave or dirty:
            filename = self.filenameForKey(key)
            with self._lock:
                if dirty:
                    value.setDirty(False)
                try:
                    sessionFile = open(filename, 'wb')
                    try:
                        try:
                            self.encoder()(value, sessionFile)
                        finally:
                            sessionFile.close()
                    except Exception:
                        # remove the session file because it is corrupt
                        os.remove(filename)
                        raise  # raise original exception
                except Exception:  # error pickling the session
                    if dirty:
                        value.setDirty()
                    print("Error saving session to disk:", key)
                    self.application().handleException()

    def __delitem__(self, key):
        """Delete a session item, removing its session file."""
        filename = self.filenameForKey(key)
        if not os.path.exists(filename):
            raise KeyError(key)
        session = self[key]
        if not session.isExpired():
            session.expiring()
        try:
            os.remove(filename)
        except Exception:
            pass

    def __contains__(self, key):
        """Check whether the session store has a file for the given key."""
        return os.path.exists(self.filenameForKey(key))

    def __iter__(self):
        """Return an iterator over the stored session keys."""
        ext = self._extension
        pos = -len(ext)
        # note that iglob is slower here, since it's based on listdir
        for filename in os.listdir(self._sessionDir):
            if filename.endswith(ext):
                yield filename[:pos]

    def removeKey(self, key):
        """Remove the session file for the given key."""
        filename = self.filenameForKey(key)
        try:
            os.remove(filename)
        except Exception:
            pass

    def keys(self):
        """Return a list with the keys of all the stored sessions."""
        return list(iter(self))

    def clear(self):
        """Clear the session file store, removing all of the session files."""
        for key in self:
            self.removeKey(key)

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        with self._lock:
            try:
                return self[key]
            except KeyError:
                self[key] = default
                return default

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        with self._lock:
            if default is NoDefault:
                value = self[key]
                self.removeKey(key)
                return value
            if key in self:
                return self[key]
            return default

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save session, writing it to the session file now."""
        if self._alwaysSave or session.isDirty():
            self[session.identifier()] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store."""
        # sessions have all been saved to files already

    # endregion Application support

    # region Self utility

    def filenameForKey(self, key):
        """Return the name of the session file for the given key."""
        return os.path.join(self._sessionDir, key + self._extension)

    # endregion Self utility



from MiscUtils.Configurable import Configurable

from .ExamplePage import ExamplePage


class SecurePage(ExamplePage, Configurable):
    """Password-based security example.

    This class is an example of how to implement username and password-based
    security using Webware. Use a SecurePage class like this one as the
    base class for any pages that you want to require login. Modify
    the isUserNameAndPassword method to perform validation in whatever
    way you desire, such as a back-end database lookup. You might also
    want to modify loginUser so that it automatically brings in additional
    information about the user and stores it in session variables.

    You can turn off security by creating a config file called
    SecurePage.config in the Configs directory with the following contents:

        RequireLogin = False

    To do: Integrate this functionality with UserKit.
    Make more of the functionality configurable in the config file.
    """

    def __init__(self):
        ExamplePage.__init__(self)
        Configurable.__init__(self)

    def awake(self, transaction):
        ExamplePage.awake(self, transaction)  # awaken our superclass
        session = transaction.session()
        request = transaction.request()
        if self.setting('RequireLogin'):
            # Handle four cases:
            # 1. login attempt
            # 2. logout
            # 3. already logged in
            # 4. not already logged in
            setField, delField = request.setField, request.delField
            app = transaction.application()
            # Are they logging out?
            if request.hasField('logout'):
                # They are logging out. Clear all session variables:
                session.values().clear()
                setField('extra', 'You have been logged out.')
                setField('action', request.urlPath().rpartition('/')[2])
                app.forward(transaction, 'LoginPage')
            # Is the session expired?
            elif request.isSessionExpired():
                # Login page with a "logged out" message.
                session.values().clear()
                setField('extra', 'Your session has expired.')
                setField('action', request.urlPath().rpartition('/')[2])
                app.forward(transaction, 'LoginPage')
            # Are they already logged in?
            elif self.loggedInUser():
                return
            # Are they logging in?
            elif (request.hasField('login')
                  and request.hasField('username')
                  and request.hasField('password')):
                # They are logging in. Get login id and clear session:
                loginId = session.value('loginId', None)
                session.values().clear()
                # Get request fields:
                username = request.field('username')
                password = request.field('password')
                # Check if they can successfully log in.
                # The loginId must match what was previously sent.
                if (request.field('loginId', 'noLogin') == loginId
                        and self.loginUser(username, password)):
                    # Successful login.
                    # Clear out the login parameters:
                    delField('username')
                    delField('password')
                    delField('login')
                    delField('loginId')
                else:
                    # Failed login attempt; have them try again:
                    setField('extra', 'Login failed. Please try again.')
                    setField('action', request.urlPath().rpartition('/')[2])
                    app.forward(transaction, 'LoginPage')
            else:
                # They need to log in.
                session.values().clear()
                # Send them to the login page:
                setField('action', request.urlPath().rpartition('/')[2])
                app.forward(transaction, 'LoginPage')
        else:
            # No login is required.
            # Are they logging out?
            if request.hasField('logout'):
                # They are logging out.  Clear all session variables.
                session.values().clear()
            # Write the page:
            ExamplePage.writeHTML(self)

    def respond(self, transaction):
        # If the user is already logged in, then process this request normally.
        # Otherwise, do nothing.
        # All of the login logic has already happened in awake().
        if self.loggedInUser():
            ExamplePage.respond(self, transaction)

    @staticmethod
    def isValidUserAndPassword(username, password):
        # Replace this with a database lookup,
        # or whatever you're using for authentication...
        users = ('alice', 'bob')
        return username.lower() in users and password == username

    def loginUser(self, username, password):
        # We mark a user as logged-in by setting a session variable
        # called authenticated_user to the logged-in username.
        # Here, you could also pull in additional information about
        # this user (such as a user ID or user preferences) and store
        # that information in session variables.
        if self.isValidUserAndPassword(username, password):
            self.session().setValue('authenticated_user', username)
            return True
        self.session().setValue('authenticated_user', None)
        return False

    def loggedInUser(self):
        # Gets the name of the logged-in user, or returns None
        # if there is no logged-in user.
        return self.session().value('authenticated_user', None)

    def defaultConfig(self):
        return dict(RequireLogin=True)

    def configFilename(self):
        return 'Configs/SecurePage.config'



import unittest

from MiscUtils.MixIn import MixIn


class TestMixIn(unittest.TestCase):

    def setUp(self):
        class Base:
            attr0 = 'this is Base.attr0'
            attr1 = 'this is Base.attr1'

            def foo(self):
                return 'this is Base.foo'

        class PyClass(Base):
            attr1 = 'this is PyClass.attr1'
            attr2 = 'this is PyClass.attr2'
            attr3 = 'this is PyClass.attr3'

            def foo(self):
                return 'this is PyClass.foo'

            @classmethod
            def cFoo(cls):
                return 'this is PyClass.cFoo'

            @staticmethod
            def sFoo():
                return 'this is PyClass.sFoo'

        class MixIn1:
            attr3 = 'this is MixIn1.attr3'
            attr4 = 'this is MixIn1.attr4'

            def foo(self):
                return 'this is MixIn1.foo'

            @classmethod
            def cFoo(cls):
                return 'this is MixIn1.cFoo'

            @staticmethod
            def sFoo():
                return 'this is MixIn1.sFoo'

        class MixIn2Base:
            attr5 = 'this is MixIn2Base.attr5'

            def bar(self):
                return 'this is MixIn2Base.bar'

            def baz(self):
                return 'this is MixIn2Base.baz'

        class MixIn2(MixIn2Base):
            attr5 = 'this is MixIn2.attr5'
            attr6 = 'this is MixIn2.attr6'

            def bar(self):
                return 'this is MixIn2.bar'

        self.base = Base
        self.pyClass = PyClass
        self.mixIn1 = MixIn1
        self.mixIn2Base = MixIn2Base
        self.mixIn2 = MixIn2

    def testMixInSameAsClass(self):
        self.assertRaisesRegex(
            TypeError, 'mixInClass is the same as pyClass',
            MixIn, self.pyClass, self.pyClass)

    # pylint: disable=no-member

    def testMakeAncestor1(self):
        cls = MixIn(self.pyClass, self.mixIn1, makeAncestor=True)
        self.assertEqual(cls.__bases__, (self.mixIn1, self.base))
        self.assertFalse(hasattr(cls, 'mixInsForPyClass'))
        self.assertFalse(hasattr(self.mixIn1, 'mixInSuperFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is PyClass.attr3')
        self.assertEqual(cls.attr4, 'this is MixIn1.attr4')
        self.assertEqual(cls.cFoo(), 'this is PyClass.cFoo')
        self.assertEqual(cls.sFoo(), 'this is PyClass.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is PyClass.foo')

    def testMakeAncestor2(self):
        cls = MixIn(self.pyClass, self.mixIn2, makeAncestor=True)
        self.assertEqual(cls.__bases__, (self.mixIn2, self.base))
        self.assertFalse(hasattr(cls, 'mixInsForPyClass'))
        self.assertFalse(hasattr(self.mixIn2, 'mixInSuperFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is PyClass.attr3')
        self.assertEqual(cls.attr5, 'this is MixIn2.attr5')
        self.assertEqual(cls.attr6, 'this is MixIn2.attr6')
        self.assertEqual(cls.cFoo(), 'this is PyClass.cFoo')
        self.assertEqual(cls.sFoo(), 'this is PyClass.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is PyClass.foo')
        self.assertEqual(obj.bar(), 'this is MixIn2.bar')
        self.assertEqual(obj.baz(), 'this is MixIn2Base.baz')

    def testMixIn1(self):
        cls = self.pyClass
        MixIn(cls, self.mixIn1)
        self.assertEqual(cls.__bases__, (self.base,))
        self.assertEqual(cls.mixInsForPyClass, [self.mixIn1])
        self.assertFalse(hasattr(self.mixIn1, 'mixInSuperFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is MixIn1.attr3')
        self.assertEqual(cls.attr4, 'this is MixIn1.attr4')
        self.assertEqual(cls.cFoo(), 'this is MixIn1.cFoo')
        self.assertEqual(cls.sFoo(), 'this is MixIn1.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is MixIn1.foo')

    def testMixIn2(self):
        cls = self.pyClass
        MixIn(cls, self.mixIn2)
        self.assertEqual(cls.__bases__, (self.base,))
        self.assertEqual(cls.mixInsForPyClass, [self.mixIn2Base, self.mixIn2])
        self.assertFalse(hasattr(self.mixIn2, 'mixInSuperFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is PyClass.attr3')
        self.assertEqual(cls.attr5, 'this is MixIn2.attr5')
        self.assertEqual(cls.attr6, 'this is MixIn2.attr6')
        self.assertEqual(cls.cFoo(), 'this is PyClass.cFoo')
        self.assertEqual(cls.sFoo(), 'this is PyClass.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is PyClass.foo')
        self.assertEqual(obj.bar(), 'this is MixIn2.bar')
        self.assertEqual(obj.baz(), 'this is MixIn2Base.baz')

    def testMixIn1And2(self):
        cls = self.pyClass
        MixIn(cls, self.mixIn1)
        MixIn(cls, self.mixIn2)
        self.assertEqual(cls.__bases__, (self.base,))
        self.assertEqual(cls.mixInsForPyClass, [
            self.mixIn1, self.mixIn2Base, self.mixIn2])
        self.assertFalse(hasattr(self.mixIn1, 'mixInSuperFoo'))
        self.assertFalse(hasattr(self.mixIn2, 'mixInSuperFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is MixIn1.attr3')
        self.assertEqual(cls.attr4, 'this is MixIn1.attr4')
        self.assertEqual(cls.attr5, 'this is MixIn2.attr5')
        self.assertEqual(cls.attr6, 'this is MixIn2.attr6')
        self.assertEqual(cls.cFoo(), 'this is MixIn1.cFoo')
        self.assertEqual(cls.sFoo(), 'this is MixIn1.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is MixIn1.foo')
        self.assertEqual(obj.bar(), 'this is MixIn2.bar')
        self.assertEqual(obj.baz(), 'this is MixIn2Base.baz')

    def testMixIn1WithSuper(self):
        cls, mixIn = self.pyClass, self.mixIn1
        MixIn(cls, mixIn, mixInSuperMethods=True)
        self.assertEqual(cls.__bases__, (self.base,))
        self.assertEqual(cls.mixInsForPyClass, [mixIn])
        self.assertTrue(hasattr(mixIn, 'mixInSuperFoo'))
        self.assertTrue(hasattr(mixIn, 'mixInSuperCFoo'))
        self.assertTrue(hasattr(mixIn, 'mixInSuperSFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is MixIn1.attr3')
        self.assertEqual(cls.attr4, 'this is MixIn1.attr4')
        self.assertEqual(cls.cFoo(), 'this is MixIn1.cFoo')
        self.assertEqual(cls.sFoo(), 'this is MixIn1.sFoo')
        self.assertEqual(mixIn.mixInSuperCFoo(), 'this is PyClass.cFoo')
        self.assertEqual(mixIn.mixInSuperSFoo(), 'this is PyClass.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is MixIn1.foo')
        superFoo = mixIn.mixInSuperFoo
        self.assertEqual(superFoo(self), 'this is PyClass.foo')

    def testMixIn2WithSuper(self):
        cls, mixIn = self.pyClass, self.mixIn2
        MixIn(cls, mixIn, mixInSuperMethods=True)
        self.assertEqual(cls.__bases__, (self.base,))
        self.assertEqual(cls.mixInsForPyClass, [self.mixIn2Base, mixIn])
        self.assertFalse(hasattr(mixIn, 'mixInSuperFoo'))
        self.assertFalse(hasattr(mixIn, 'mixInSuperCFoo'))
        self.assertFalse(hasattr(mixIn, 'mixInSuperSFoo'))
        self.assertEqual(cls.attr0, 'this is Base.attr0')
        self.assertEqual(cls.attr1, 'this is PyClass.attr1')
        self.assertEqual(cls.attr2, 'this is PyClass.attr2')
        self.assertEqual(cls.attr3, 'this is PyClass.attr3')
        self.assertEqual(cls.attr5, 'this is MixIn2.attr5')
        self.assertEqual(cls.attr6, 'this is MixIn2.attr6')
        self.assertEqual(cls.cFoo(), 'this is PyClass.cFoo')
        self.assertEqual(cls.sFoo(), 'this is PyClass.sFoo')
        obj = cls()
        self.assertEqual(obj.foo(), 'this is PyClass.foo')
        self.assertEqual(obj.bar(), 'this is MixIn2.bar')
        self.assertEqual(obj.baz(), 'this is MixIn2Base.baz')



import unittest

from MiscUtils import AbstractError, NoDefault
from MiscUtils.NamedValueAccess import (
    NamedValueAccessError, valueForKey, valueForName)


class T:  # pylint: disable=invalid-name
    pass


class T1(T):

    def foo(self):
        return 1


class T2(T):

    def _foo(self):
        return 1


class T3(T):

    def foo(self):
        return 1

    def _foo(self):
        return 0


class T4(T):

    def foo(self):
        return 1

    def __init__(self):
        self._foo = 0


class T5(T):

    def __init__(self):
        self.foo = 0

    def _foo(self):
        return 1


class T6(T):

    def __init__(self):
        self.foo = 1
        self._foo = 0


# Make a list of all the 'T' classes which are used in testing
tClasses = [globals()[name] for name in dir()
            if name.startswith('T') and name[1:].isdigit()]


class LookupTest:
    """Abstract super class for the test cases covering the functions.

    Subclasses must implement self.lookup() and can make use of
    self.classes and self.objs.
    """

    def setUp(self):
        self.setUpClasses()
        self.setUpObjects()

    def setUpClasses(self):
        self.classes = tClasses

    def setUpObjects(self):
        self.objs = [klass() for klass in self.classes]

    def lookup(self, obj, key, default=NoDefault):
        raise AbstractError(self.__class__)

    def testBasicAccess(self):
        """Check the basic access functionality.

        Invoke the look up function with key 'foo', expecting 1 in return.
        Invoke the look up with 'bar', expected an exception.
        Invoke the look up with 'bar' and default 2, expecting 2.
        """
        func = self.lookup
        # pylint: disable=assignment-from-no-return,no-member
        for obj in self.objs:
            value = func(obj, 'foo')
            self.assertEqual(value, 1, 'value = %r, obj = %r' % (value, obj))
            self.assertRaises(NamedValueAccessError, func, obj, 'bar')
            value = func(obj, 'bar', 2)
            self.assertEqual(value, 2, 'value = %r, obj = %r' % (value, obj))

    def testBasicAccessRepeated(self):
        """Just repeat checkBasicAccess multiple times to check stability."""
        for _count in range(50):
            # Yes, it's safe to invoke this other particular test
            # multiple times without the usual setUp()/tearDown() cycle
            self.testBasicAccess()


class ValueForKeyTest(LookupTest, unittest.TestCase):

    def lookup(self, obj, key, default=NoDefault):
        return valueForKey(obj, key, default)


class ValueForNameTest(LookupTest, unittest.TestCase):

    def lookup(self, obj, key, default=NoDefault):
        return valueForName(obj, key, default)

    def testNamedValueAccess(self):
        objs = self.objs

        # link the objects
        for i in range(len(objs)-1):
            objs[i].nextObject = objs[i+1]

        # test the links
        for i in range(len(objs)):
            name = 'nextObject.' * i + 'foo'
            self.assertEqual(self.lookup(objs[0], name), 1)

    def testDicts(self):
        d = {'origin': {'x': 1, 'y': 2},
             'size': {'width': 3, 'height': 4}}
        obj = self.objs[0]
        obj.rect = d

        self.assertEqual(self.lookup(d, 'origin.x'), 1)
        self.assertTrue(self.lookup(obj, 'rect.origin.x'))

        self.assertRaises(NamedValueAccessError, self.lookup, d, 'bar')
        self.assertRaises(NamedValueAccessError, self.lookup, obj, 'bar')

        self.assertEqual(self.lookup(d, 'bar', 2), 2)
        self.assertEqual(self.lookup(obj, 'rect.bar', 2), 2)



"""DateParser.py

Convert string representations of dates to Python datetime objects.

If installed, we will use the python-dateutil package to parse dates,
otherwise we try to use the strptime function in the Python standard library
with several frequently used formats.
"""

__all__ = ['parseDateTime', 'parseDate', 'parseTime']

try:

    from dateutil.parser import parse as parseDateTime

except ImportError:  # dateutil not available

    from datetime import datetime

    strpdatetime = datetime.strptime

    def parseDateTime(s):
        """Return a datetime object corresponding to the given string."""
        formats = (
            "%a %b %d %H:%M:%S %Y", "%a, %d-%b-%Y %H:%M:%S",
            "%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S",
            "%Y%m%d %H:%M:%S", "%Y%m%dT%H:%M:%S",
            "%Y%m%d %H%M%S", "%Y%m%dT%H%M%S",
            "%m/%d/%y %H:%M:%S", "%Y-%m-%d %H:%M",
            "%Y-%m-%d", "%Y%m%d", "%m/%d/%y",
            "%H:%M:%S", "%H:%M", "%c")
        for fmt in formats:
            try:
                return strpdatetime(s, fmt)
            except ValueError:
                pass
        raise ValueError(f'Cannot parse date/time {s}')


def parseDate(s):
    """Return a date object corresponding to the given string."""
    return parseDateTime(s).date()


def parseTime(s):
    """Return a time object corresponding to the given string."""
    return parseDateTime(s).time()



"""Session store using the Redis in-memory data store."""

from pickle import loads, dumps

try:
    import redis  # pylint: disable=import-error
except Exception as e:
    raise ImportError(
        "For using Redis sessions,"
        " redis-py must be installed.") from e

from MiscUtils import NoDefault

from SessionStore import SessionStore

debug = False


class SessionRedisStore(SessionStore):
    """A session store using Redis.

    Stores the sessions in a single Redis store using 'last write wins'
    semantics. This increases fault tolerance and allows server clustering.
    In clustering configurations with concurrent writes for the same
    session(s) the last writer will always overwrite the session.

    The keys are prefixed with a configurable namespace, allowing you to
    store other data in the same Redis system.

    Cleaning/timing out of sessions is performed by Redis itself since
    no single application can know about the existence of all sessions or
    the last access for a given session. Besides it is built in Redis
    functionality. Consequently, correct sizing of Redis is necessary
    to hold all user's session data.

    You need to install the redis client to be able to use this module:
    https://pypi.python.org/pypi/redis
    You also need a Redis server: https://redis.io/

    Contributed by Christoph Zwerschke, August 2016.
    """

    # region Init

    def __init__(self, app):
        SessionStore.__init__(self, app)

        # the list of redis servers
        self._host = app.setting('RedisHost', 'localhost')
        self._port = app.setting('RedisPort', 6379)
        self._db = app.setting('RedisDb', 0)
        self._password = app.setting('RedisPassword', None)

        # timeout in seconds
        self._sessionTimeout = app.setting('SessionTimeout', 180) * 60

        # the redis "namespace" used by our store
        self._namespace = app.setting(
            'RedisNamespace', 'WebwareSession:') or ''

        self._redis = redis.StrictRedis(
            self._host, self._port, self._db, self._password)

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions in the store."""
        if debug:
            print(">> len()")
        return len(self.keys())

    def __getitem__(self, key):
        """Get a session item, reading it from the store."""
        if debug:
            print(f">> getitem({key})")
        # returns None if key non-existent or no server to contact
        try:
            value = loads(self._redis.get(self.redisKey(key)))
        except Exception:
            value = None
        if value is None:
            # SessionStore expects KeyError when no result
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        """Set a session item, writing it to the store."""
        if debug:
            print(f">> setitem({key}, {value})")
        dirty = value.isDirty()
        if self._alwaysSave or dirty:
            if dirty:
                value.setDirty(False)
            try:
                self._redis.setex(
                    self.redisKey(key), self._sessionTimeout, dumps(value, -1))
            except Exception as exc:
                if dirty:
                    value.setDirty()
                # Not able to store the session is a failure
                print(f"Error saving session {key!r} to redis: {exc}")
                self.application().handleException()

    def __delitem__(self, key):
        """Delete a session item from the store.

        Note that in contracts with SessionFileStore,
        not finding a key to delete isn't a KeyError.
        """
        if debug:
            print(f">> delitem({key})")
        session = self[key]
        if not session.isExpired():
            session.expiring()
        try:
            self._redis.delete(self.redisKey(key))
        except Exception as exc:
            # Not able to delete the session is a failure
            print(f"Error deleting session {key!r} from redis: {exc}")
            self.application().handleException()

    def __contains__(self, key):
        """Check whether the session store has a given key."""
        if debug:
            print(f">> contains({key})")
        try:
            return self._redis.exists(self.redisKey(key))
        except Exception as exc:
            # Not able to check the session is a failure
            print(f"Error checking session {key!r} from redis: {exc}")
            self.application().handleException()

    def __iter__(self):
        """Return an iterator over the stored session keys."""
        if debug:
            print(">> iter()")
        return self.keys().__iter__()

    def keys(self):
        """Return a list with the keys of all the stored sessions."""
        if debug:
            print(">> keys()")
        try:
            if self._namespace:
                n = len(self._namespace)
                return [k[n:] for k in self._redis.keys(self.redisKey('*'))]
            return self._redis.keys(self.redisKey('*'))
        except Exception as exc:
            # Not able to get the keys is a failure
            print("Error checking sessions from redis:", exc)
            self.application().handleException()

    def clear(self):
        """Clear the session store, removing all of its items."""
        if debug:
            print(">> clear()")
        try:
            if self._namespace:
                self._redis.delete(*self._redis.keys(self.redisKey('*')))
            else:
                self._redis.flushdb()
        except Exception as exc:
            # Not able to clear the store is a failure
            print("Error clearing sessions from redis:", exc)
            self.application().handleException()

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        if debug:
            print(f">> setdefault({key}, {default})")
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        if debug:
            print(f">> pop({key}, {default})")
        if default is NoDefault:
            value = self[key]
            del self[key]
            return value
        try:
            value = self[key]
        except KeyError:
            return default
        else:
            del self[key]
            return value

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save potentially changed session in the store."""
        if debug:
            print(f">> storeSession({session})")
        self[session.identifier()] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store.

        Should be used (only) when the application server is shut down.
        This closes the connections to the Redis server.
        """
        if debug:
            print(">> storeAllSessions()")
        try:
            self._redis.connection_pool.disconnect()
        except Exception as exc:
            print("Not able to disconnect from redis:", exc)

    def cleanStaleSessions(self, _task=None):
        """Clean stale sessions.

        Redis does this on its own, so we do nothing here.
        """
        if debug:
            print(">> cleanStaleSessions()")

    # endregion Application support

    # region Auxiliary methods

    def redisKey(self, key):
        """Create the real key with namespace to be used with Redis."""
        return self._namespace + key

    # endregion Auxiliary methods



#
# JSON-RPC example servlet contributed by Jean-Francois Pieronne
#

from JSONRPCServlet import JSONRPCServlet


class JSONRPCExample(JSONRPCServlet):
    """Example JSON-RPC servlet.

    To try it out, use the JSONRPCClient servlet.
    """

    def __init__(self):
        JSONRPCServlet.__init__(self)

    @staticmethod
    def echo(msg):
        return msg

    @staticmethod
    def reverse(msg):
        return msg[::-1]

    @staticmethod
    def uppercase(msg):
        return msg.upper()

    @staticmethod
    def lowercase(msg):
        return msg.lower()

    @staticmethod
    def exposedMethods():
        return ['echo', 'reverse', 'uppercase', 'lowercase']



"""Content producing HTTP servlet."""

from WebUtils.Funcs import urlDecode, urlEncode
from HTTPServlet import HTTPServlet
from Application import EndResponse


class HTTPContentError(Exception):
    """HTTP content error"""


class HTTPContent(HTTPServlet):
    """Content producing HTTP servlet.

    HTTPContent is a type of HTTPServlet that is more convenient for
    Servlets which represent content generated in response to
    GET and POST requests.  If you are generating HTML content, you
    you probably want your servlet to inherit from Page, which contains
    many HTML-related convenience methods.

    If you are generating non-HTML content, it is appropriate to inherit
    from this class directly.

    Subclasses typically override defaultAction().

    In `awake`, the page sets self attributes: `_transaction`, `_response`
    and `_request` which subclasses should use as appropriate.

    For the purposes of output, the `write` and `writeln`
    convenience methods are provided.

    If you plan to produce HTML content, you should start by looking
    at Page instead of this lower-level class.
    """

    # region Transactions

    def awake(self, transaction):
        """Let servlet awake.

        Makes instance variables from the transaction. This is where Page
        becomes unthreadsafe, as the page is tied to the transaction.
        This is also what allows us to implement functions like `write`,
        where you don't need to pass in the transaction or response.
        """
        HTTPServlet.awake(self, transaction)
        self._response = transaction.response()
        self._request = transaction.request()
        self._session = None  # don't create unless needed

    def respondToGet(self, transaction):
        """Respond to GET.

        Invoked in response to a GET request method. All methods
        are passed to `_respond`.
        """
        self._respond(transaction)

    def respondToPost(self, transaction):
        """Respond to POST.

        Invoked in response to a POST request method. All methods
        are passed to `_respond`.
        """
        self._respond(transaction)

    def _respond(self, transaction):
        """Respond to action.

        Handles actions if an ``_action_`` or ``_action_name`` field is
        defined, otherwise invokes `writeHTML`. This implementation makes
        sure that exactly one action per request is handled. ``_action_``
        takes precedence over ``_action_name``; and if there are multiple
        ``action_name`` fields, the precedence is given by the order of
        the names in the actions() method. If no action field matches,
        the default action is run. The value of the ``_action_`` field
        is transformed to a method name using the methodNameForAction(),
        whereas ``name`` in ``_action_name`` is left unchanged.

        Invoked by both `respondToGet` and `respondToPost`.
        """
        req = transaction.request()
        prefix = self._actionPrefix
        if prefix:
            # First check whether there is an _action_ field:
            if req.hasField(prefix):
                action = self.methodNameForAction(req.field(prefix))
                if action in self.actions():
                    self.handleAction(action)
                    return
            # Next, check whether there is an _action_name field:
            for action in self.actions():
                name = prefix + action
                if req.hasField(name) or (
                        req.hasField(name + '.x')
                        and req.hasField(name + '.y')):
                    self.handleAction(action)
                    return
            # If no action was found, run the default:
        self.defaultAction()

    def defaultAction(self):
        """Default action.

        The core method that gets called as a result of requests.
        Subclasses should override this.
        """
        # base method does nothing

    def sleep(self, transaction):
        """Let servlet sleep again.

        We unset some variables. Very boring.
        """
        self._session = None
        self._request = None
        self._response = None
        self._transaction = None
        HTTPServlet.sleep(self, transaction)

    # endregion Transactions

    # region Access

    def application(self):
        """The `Application` instance we're using."""
        return self._transaction.application()

    def transaction(self):
        """The `Transaction` we're currently handling."""
        return self._transaction

    def request(self):
        """The request (`HTTPRequest`) we're handling."""
        return self._request

    def response(self):
        """The response (`HTTPResponse`) we're handling."""
        return self._response

    def session(self):
        """The session object.

        This provides a state for the current user
        (associated with a browser instance, really).
        If no session exists, then a session will be created.
        """
        if not self._session:
            self._session = self._transaction.session()
        return self._session

    # endregion Access

    # region Writing

    def write(self, *args):
        """Write to output.

        Writes the arguments, which are turned to strings (with `str`)
        and concatenated before being written to the response.
        Unicode strings must be encoded before they can be written.
        """
        for arg in args:
            self._response.write(arg)

    def writeln(self, *args):
        """Write to output with newline.

        Writes the arguments (like `write`), adding a newline after.
        Unicode strings must be encoded before they can be written.
        """
        for arg in args:
            self._response.write(arg)
        self._response.write('\n')

    def outputEncoding(self):
        """Get the default output encoding of the application."""
        return self.application().outputEncoding()

    # endregion Writing

    # region Threading

    def canBeThreaded(self):
        """Declares whether servlet can be threaded.

        Returns False because of the instance variables we set up in `awake`.
        """
        return False

    # endregion Threading

    # region Actions

    _actionPrefix = '_action_'

    def handleAction(self, action):
        """Handle action.

        Invoked by `_respond` when a legitimate action has
        been found in a form. Invokes `preAction`, the actual
        action method and `postAction`.

        Subclasses rarely override this method.
        """
        self.preAction(action)
        getattr(self, action)()
        self.postAction(action)

    def actions(self):
        """The allowed actions.

        Returns a list or a set of method names that are allowable
        actions from HTML forms. The default implementation returns [].
        See `_respond` for more about actions.
        """
        return []

    def preAction(self, actionName):
        """Things to do before action.

        Invoked by self prior to invoking a action method.
        The `actionName` is passed to this method,
        although it seems a generally bad idea to rely on this.
        However, it's still provided just in case you need that hook.

        By default this does nothing.
        """
        # base method does nothing

    def postAction(self, actionName):
        """Things to do after action.

        Invoked by self after invoking a action method.
        Subclasses may override to customize and may or may not
        invoke super as they see fit.
        The `actionName` is passed to this method,
        although it seems a generally bad idea to rely on this.
        However, it's still provided just in case you need that hook.

        By default this does nothing.
        """
        # base method does nothing

    def methodNameForAction(self, name):
        """Return method name for an action name.

        Invoked by _respond() to determine the method name for a given action
        name which has been derived as the value of an ``_action_`` field.
        Since this is usually the label of an HTML submit button in a form,
        it is often needed to transform it in order to get a valid method name
        (for instance, blanks could be replaced by underscores and the like).
        This default implementation of the name transformation is the identity,
        it simply returns the name. Subclasses should override this method
        when action names don't match their method names; they could "mangle"
        the action names or look the method names up in a dictionary.
        """
        return name

    @staticmethod
    def urlEncode(s):
        """Quotes special characters using the % substitutions.

        This method does the same as the `urllib.quote_plus()` function.
        """
        return urlEncode(s)

    @staticmethod
    def urlDecode(s):
        """Turn special % characters into actual characters.

        This method does the same as the `urllib.unquote_plus()` function.
        """
        return urlDecode(s)

    def forward(self, url):
        """Forward request.

        Forwards this request to another servlet.
        See `Application.forward` for details.
        The main difference is that here you don't have
        to pass in the transaction as the first argument.
        """
        self.application().forward(self.transaction(), url)

    def includeURL(self, url):
        """Include output from other servlet.

        Includes the response of another servlet
        in the current servlet's response.
        See `Application.includeURL` for details.
        The main difference is that here you don't have
        to pass in the transaction as the first argument.
        """
        self.application().includeURL(self.transaction(), url)

    def callMethodOfServlet(self, url, method, *args, **kwargs):
        """Call a method of another servlet.

        See `Application.callMethodOfServlet` for details.
        The main difference is that here you don't have
        to pass in the transaction as the first argument.
        """
        return self.application().callMethodOfServlet(
            self.transaction(), url, method, *args, **kwargs)

    @staticmethod
    def endResponse():
        """End response.

        When this method is called during `awake` or `respond`,
        servlet processing will end immediately,
        and the accumulated response will be sent.

        Note that `sleep` will still be called, providing a
        chance to clean up or free any resources.
        """
        raise EndResponse

    def sendRedirectAndEnd(self, url, status=None):
        """Send redirect and end.

        Sends a redirect back to the client and ends the response.
        This is a very popular pattern.
        """
        self.response().sendRedirect(url, status)
        self.endResponse()

    def sendRedirectPermanentAndEnd(self, url):
        """Send permanent redirect and end."""
        self.response().sendRedirectPermanent(url)
        self.endResponse()

    def sendRedirectSeeOtherAndEnd(self, url):
        """Send redirect to a URL to be retrieved with GET and end.

        This is the proper method for the Post/Redirect/Get pattern.
        """
        self.response().sendRedirectSeeOther(url)
        self.endResponse()

    def sendRedirectTemporaryAndEnd(self, url):
        """Send temporary redirect and end."""
        self.response().sendRedirectTemporary(url)
        self.endResponse()

    # endregion Actions

    # region Utility

    def sessionEncode(self, url=None):
        """Utility function to access `Session.sessionEncode`.

        Takes a url and adds the session ID as a parameter.
        This is for cases where you don't know if the client
        will accepts cookies.
        """
        if url is None:
            url = self.request().uri()
        return self.session().sessionEncode(url)

    # endregion Utility

    # region Exception Reports

    def writeExceptionReport(self, handler):
        """Write extra information to the exception report.

        The `handler` argument is the exception handler, and information
        is written there (using `writeTitle`, `write`, and `writeln`).
        This information is added to the exception report.

        See `ExceptionHandler` for more information.
        """
        handler.writeln('''
<p>Servlets can provide debugging information here by overriding
<code>writeExceptionReport()</code>.</p><p>For example:</p>
<pre>
exceptionReportAttrs = ['foo', 'bar', 'baz']
def writeExceptionReport(self, handler):
    handler.writeTitle(self.__class__.__name__)
    handler.writeAttrs(self, self.exceptionReportAttrs)
    handler.write('any string')
</pre>
<p>See ExceptionHandler.py for more information.</p>
''')

    # endregion Exception Reports
